# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class for KFP Registry Client."""

import json
import logging
import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union

import google.auth
from google.auth import credentials
import requests

_KNOWN_HOSTS_REGEX = {
    'kfp_pkg_dev': (
        r'(^https\:\/\/(?P<location>[\w\-]+)\-kfp\.pkg\.dev\/(?P<project_id>.*)\/(?P<repo_id>.*))',
        os.path.join(os.path.dirname(__file__), 'context/kfp_pkg_dev.json'))
}

_DEFAULT_JSON_HEADER = {
    'Content-type': 'application/json',
}

_VERSION_PREFIX = 'sha256:'

LOCAL_REGISTRY_CREDENTIAL = os.path.expanduser(
    '~/.config/kfp/registry_credentials.json')
LOCAL_REGISTRY_CONTEXT = os.path.expanduser(
    '~/.config/kfp/registry_context.json')
DEFAULT_REGISTRY_CONTEXT = os.path.join(
    os.path.dirname(__file__), 'context/default_pkg_dev.json')


class _SafeDict(dict):
    """Class for safely handling missing keys in .format_map."""

    def __missing__(self, key: str) -> str:
        """Handle missing keys by adding them back.

        Args:
            key: The key itself.

        Returns:
            The key with curly braces.
        """
        return '{' + key + '}'


class ApiAuth(requests.auth.AuthBase):
    """Class for registry authentication using an API token.

    Args:
        token: The API token.

    Example:
      ::

        client = RegistryClient(
            host='https://us-central1-kfp.pkg.dev/proj/repo', auth=ApiAuth('my_token'))
    """

    def __init__(self, token: str) -> None:
        """Initializes the ApiAuth object."""
        self._token = token

    def __call__(self,
                 request: requests.PreparedRequest) -> requests.PreparedRequest:
        request.headers['authorization'] = 'Bearer ' + self._token
        return request


class RegistryClient:
    """Class for communicating with registry hosts.

    Args:
        host: The address of the registry host. The host needs to be specified here or in the config file.
        auth: Authentication using ``requests.auth.AuthBase`` or ``google.auth.credentials.Credentials``.
        config_file: The location of the local config file. If not specified, defaults to ``'~/.config/kfp/context.json'`` (if it exists).
        auth_file: The location of the local config file that contains the authentication token. If not specified, defaults to ``'~/.config/kfp/registry_credentials.json'`` (if it exists).
    """

    def __init__(self,
                 host: Optional[str],
                 auth: Optional[Union[requests.auth.AuthBase,
                                      credentials.Credentials]] = None,
                 config_file: Optional[str] = None,
                 auth_file: Optional[str] = None) -> None:
        """Initializes the RegistryClient."""
        self._host = ''
        self._known_host_key = ''
        self._config = self._load_config(host, config_file)
        self._auth = self._load_auth(auth, auth_file)

    def _request(self,
                 request_url: str,
                 request_body: Optional[str] = '',
                 http_request: Optional[str] = None,
                 extra_headers: Optional[dict] = None) -> requests.Response:
        """Calls the HTTP request.

        Args:
            request_url: The address of the API endpoint to send the request to.
            request_body: Body of the request.
            http_request: Type of HTTP request (post, get, delete etc, defaults to get).
            extra_headers: Any extra headers required.

        Returns:
            Response from the request.
        """
        self._refresh_creds()
        auth = self._get_auth()
        http_request = http_request or 'get'
        http_request_fn = getattr(requests, http_request)

        response = http_request_fn(
            url=request_url,
            data=request_body,
            headers=extra_headers,
            auth=auth)
        response.raise_for_status()

        return response

    def _is_ar_host(self) -> bool:
        """Checks if the host is on Artifact Registry.

        Returns:
            Whether the host is on Artifact Registry.
        """
        # TODO: Handle multiple known hosts.
        return self._known_host_key == 'kfp_pkg_dev'

    def _is_known_host(self) -> bool:
        """Checks if the host is a known host.

        Returns:
            Whether the host is a known host.
        """
        return bool(self._known_host_key)

    def _validate_version(self, version: str) -> None:
        """Validates the version.

        Args:
            version: Version of the package.
        """
        if not version.startswith(_VERSION_PREFIX):
            raise ValueError('Version should start with \"sha256:\".')

    def _validate_tag(self, tag: str) -> None:
        """Validates the tag.

        Args:
            tag: Tag attached to the package.
        """
        if tag.startswith(_VERSION_PREFIX):
            raise ValueError('Tag should not start with \"sha256:\".')

    def _load_auth(
        self,
        auth: Optional[Union[requests.auth.AuthBase,
                             credentials.Credentials]] = None,
        auth_file: Optional[str] = None
    ) -> Optional[Union[requests.auth.AuthBase, credentials.Credentials]]:
        """Loads the credentials for authentication.

        Args:
            auth: Authentication using ``requests.auth.AuthBase`` or ``google.auth.credentials.Credentials``.
            auth_file: The location of the local config file that contains the
                authentication token. If not specified, defaults to
                ``'~/.config/kfp/registry_credentials.json'`` (if it exists).

        Returns:
            The loaded authentication token.
        """
        if auth:
            return auth
        elif self._is_ar_host():
            auth, _ = google.auth.default()
            return auth
        elif auth_file:
            if os.path.exists(auth_file):
                # Fetch auth token using the locally stored credentials.
                with open(auth_file, 'r') as f:
                    auth_token = json.load(f)
                    return ApiAuth(auth_token)
            else:
                raise ValueError(f'Auth file not found: {auth_file}.')
        elif os.path.exists(LOCAL_REGISTRY_CREDENTIAL):
            # Fetch auth token using the locally stored credentials.
            with open(LOCAL_REGISTRY_CREDENTIAL, 'r') as f:
                auth_token = json.load(f)
                return ApiAuth(auth_token)
        return None

    def _load_config(self, host: Optional[str],
                     config_file: Optional[str]) -> dict:
        """Loads the config.

        Args:
            host: The address of the registry host.
            config_file: The location of the local config file. If not specified,
                defaults to ``'~/.config/kfp/context.json'`` (if it exists).

        Returns:
            The loaded config.
        """
        if host:
            self._host = host.rstrip('/')
        else:
            # Check config file exists
            config_file_path = ''
            if config_file:
                if os.path.exists(config_file):
                    config_file_path = config_file
                else:
                    raise ValueError(f'Config file not found: {config_file}.')
            elif os.path.exists(LOCAL_REGISTRY_CONTEXT):
                config_file_path = LOCAL_REGISTRY_CONTEXT
            # Try loading host from config file
            if config_file_path:
                with open(config_file_path, 'r') as f:
                    data = json.load(f)
                    if 'host' in data:
                        self._host = data['host']
        if not self._host:
            raise ValueError('No host found.')

        # Check if it's a known host
        for key in _KNOWN_HOSTS_REGEX.keys():
            if re.match(_KNOWN_HOSTS_REGEX[key][0], self._host):
                self._known_host_key = key
                break

        # Try loading config from known contexts or local context
        if self._is_known_host():
            config = self._load_context(
                _KNOWN_HOSTS_REGEX[self._known_host_key][1])
        elif os.path.exists(LOCAL_REGISTRY_CONTEXT):
            config = self._load_context(LOCAL_REGISTRY_CONTEXT)
        else:
            config = self._load_context(DEFAULT_REGISTRY_CONTEXT)

        # If config file is specified, add/override any extra context info needed
        if config_file and os.path.exists(config_file):
            config = self._load_context(config_file, config)

        matched = None
        if self._is_known_host():
            matched = re.match(_KNOWN_HOSTS_REGEX[self._known_host_key][0],
                               self._host)
        elif 'regex' in config:
            matched = re.match(config['regex'], self._host)

        if matched:
            map_dict = _SafeDict(**matched.groupdict(), host=self._host)
        else:
            map_dict = _SafeDict(host=self._host)

        # Replace all currently known variables with values
        for config_key in config:
            config[config_key] = config[config_key].format_map(map_dict)

        return config

    def _load_context(self,
                      config_file: str,
                      config: Optional[dict] = None) -> dict:
        """Loads the context from the given config_file.

        Args:
            config_file: The location of the config file.
            config: An existing config to set as the default config.

        Returns:
            The loaded config.
        """
        if not os.path.exists(config_file):
            raise ValueError(f'Config file not found: {config_file}.')
        with open(config_file, 'r') as f:
            loaded_config = json.load(f)
        if config:
            config.update(loaded_config)
            return config
        return loaded_config

    def _get_auth(
        self
    ) -> Optional[Union[requests.auth.AuthBase, credentials.Credentials]]:
        """Helper function to convert google credentials to AuthBase class if
        needed.

        Returns:
            An instance of the AuthBase class
        """
        auth = self._auth
        if isinstance(auth, credentials.Credentials):
            auth = ApiAuth(auth.token)
        return auth

    def _refresh_creds(self) -> None:
        """Helper function to refresh google credentials if needed."""
        if self._is_ar_host() and isinstance(
                self._auth, credentials.Credentials) and not self._auth.valid:
            self._auth.refresh(google.auth.transport.requests.Request())

    def upload_pipeline(
            self,
            file_name: str,
            tags: Optional[Union[str, List[str]]] = None,
            extra_headers: Optional[dict] = None) -> Tuple[str, str]:
        """Uploads the pipeline.

        Args:
            file_name: The name of the file to be uploaded.
            tags: Tags to be attached to the uploaded pipeline.
            extra_headers: Any extra headers required.

        Returns:
            A tuple of the package name and the version.
        """
        url = self._config['upload_url']
        self._refresh_creds()
        auth = self._get_auth()
        request_body = {}
        if tags:
            if isinstance(tags, str):
                request_body = {'tags': tags}
            elif isinstance(tags, List):
                request_body = {'tags': ','.join(tags)}

        with open(file_name, 'rb') as f:
            files = {'content': f}
            response = requests.post(
                url=url,
                data=request_body,
                headers=extra_headers,
                files=files,
                auth=auth)
        response.raise_for_status()
        [package_name, version] = response.text.split('/')

        return (package_name, version)

    def _get_download_url(self,
                          package_name: str,
                          version: Optional[str] = None,
                          tag: Optional[str] = None) -> str:
        """Gets the download url based on version or tag (either one must be
        specified).

        Args:
            package_name: Name of the package.
            version: Version of the package.
            tag: Tag attached to the package.

        Returns:
            A url for downloading the package.
        """
        if (not version) and (not tag):
            raise ValueError('Either version or tag must be specified.')
        if version:
            self._validate_version(version)
            url = self._config['download_version_url'].format(
                package_name=package_name, version=version)
        if tag:
            if version:
                logging.info(
                    'Both version and tag are specified, using version only.')
            else:
                self._validate_tag(tag)
                url = self._config['download_tag_url'].format(
                    package_name=package_name, tag=tag)
        return url

    def download_pipeline(self,
                          package_name: str,
                          version: Optional[str] = None,
                          tag: Optional[str] = None,
                          file_name: Optional[str] = None) -> str:
        """Downloads a pipeline. Either version or tag must be specified.

        Args:
            package_name: Name of the package.
            version: Version of the package.
            tag: Tag attached to the package.
            file_name: File name to be saved as. If not specified, the
                file name will be based on the package name and version/tag.

        Returns:
            The file name of the downloaded pipeline.
        """
        url = self._get_download_url(package_name, version, tag)
        response = self._request(request_url=url)

        if not file_name:
            file_name = package_name + '_'
            if version:
                self._validate_version(version)
                file_name += version[len(_VERSION_PREFIX):]
            elif tag:
                self._validate_tag(tag)
                file_name += tag
            file_name += '.yaml'

        with open(file_name, 'wb') as f:
            f.write(response.content)

        return file_name

    def get_package(self, package_name: str) -> Dict[str, Any]:
        """Gets package metadata.

        Args:
            package_name: Name of the package.

        Returns:
            The package metadata.
        """
        url = self._config['get_package_url'].format(package_name=package_name)
        response = self._request(request_url=url)

        return response.json()

    def list_packages(self) -> List[dict]:
        """Lists packages.

        Returns:
            List of packages in the repository.
        """
        url = self._config['list_packages_url']
        response = self._request(request_url=url)
        response_json = response.json()

        return response_json.get('packages', {})

    def delete_package(self, package_name: str) -> bool:
        """Deletes a package.

        Args:
            package_name: Name of the package.

        Returns:
            Whether the package was deleted successfully.
        """
        url = self._config['delete_package_url'].format(
            package_name=package_name)
        response = self._request(request_url=url, http_request='delete')
        response_json = response.json()

        return response_json['done']

    def get_version(self, package_name: str, version: str) -> Dict[str, Any]:
        """Gets package version metadata.

        Args:
            package_name: Name of the package.
            version: Version of the package.

        Returns:
            The version metadata.
        """
        self._validate_version(version)
        url = self._config['get_version_url'].format(
            package_name=package_name, version=version)
        response = self._request(request_url=url)

        return response.json()

    def list_versions(self, package_name: str) -> List[dict]:
        """Lists package versions.

        Args:
            package_name: Name of the package.

        Returns:
            List of package versions.
        """
        url = self._config['list_versions_url'].format(
            package_name=package_name)
        response = self._request(request_url=url)
        response_json = response.json()

        return response_json.get('versions', {})

    def delete_version(self, package_name: str, version: str) -> bool:
        """Deletes package version.

        Args:
            package_name: Name of the package.
            version: Version of the package.

        Returns:
            Whether the version was deleted successfully.
        """
        self._validate_version(version)
        url = self._config['delete_version_url'].format(
            package_name=package_name, version=version)
        response = self._request(request_url=url, http_request='delete')
        response_json = response.json()

        return response_json['done']

    def create_tag(self, package_name: str, version: str,
                   tag: str) -> Dict[str, Any]:
        """Creates a tag on a package version.

        Args:
            package_name: Name of the package.
            version: Version of the package.
            tag: Tag to be attached to the package version.

        Returns:
            Metadata for the created tag.
        """
        self._validate_version(version)
        self._validate_tag(tag)
        url = self._config['create_tag_url'].format(
            package_name=package_name, tag=tag)
        new_tag = {
            'name':
                '',
            'version':
                self._config['version_format'].format(
                    package_name=package_name, version=version)
        }
        response = self._request(
            request_url=url,
            request_body=json.dumps(new_tag),
            http_request='post',
            extra_headers=_DEFAULT_JSON_HEADER)

        return response.json()

    def get_tag(self, package_name: str, tag: str) -> Dict[str, Any]:
        """Gets tag metadata.

        Args:
            package_name: Name of the package.
            tag: Tag attached to the package version.

        Returns:
            The metadata for the tag.
        """
        self._validate_tag(tag)
        url = self._config['get_tag_url'].format(
            package_name=package_name, tag=tag)
        response = self._request(request_url=url)

        return response.json()

    def update_tag(self, package_name: str, version: str,
                   tag: str) -> Dict[str, Any]:
        """Updates a tag to another package version.

        Args:
            package_name: Name of the package.
            version: Version of the package.
            tag: Tag to be attached to the new package version.

        Returns:
            The metadata for the updated tag.
        """
        self._validate_version(version)
        self._validate_tag(tag)
        url = self._config['update_tag_url'].format(
            package_name=package_name, tag=tag)
        new_tag = {
            'name':
                '',
            'version':
                self._config['version_format'].format(
                    package_name=package_name, version=version)
        }
        response = self._request(
            request_url=url,
            request_body=json.dumps(new_tag),
            http_request='patch',
            extra_headers=_DEFAULT_JSON_HEADER)

        return response.json()

    def list_tags(self, package_name: str) -> List[dict]:
        """Lists package tags.

        Args:
            package_name: Name of the package.

        Returns:
            List of tags.
        """
        url = self._config['list_tags_url'].format(package_name=package_name)
        response = self._request(request_url=url)
        response_json = response.json()

        return response_json.get('tags', {})

    def delete_tag(self, package_name: str, tag: str) -> Dict[str, Any]:
        """Deletes package tag.

        Args:
            package_name: Name of the package.
            tag: Tag to be deleted.

        Returns:
            Response from the delete request.
        """
        self._validate_tag(tag)
        url = self._config['delete_tag_url'].format(
            package_name=package_name, tag=tag)
        response = self._request(request_url=url, http_request='delete')

        return response.json()



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import tempfile
import logging
import shutil
from collections import OrderedDict
from typing import Callable, Dict, List, Optional

from deprecated.sphinx import deprecated

from ..components._components import _create_task_factory_from_component_spec
from ..components._python_op import _func_to_component_spec
from ._container_builder import ContainerBuilder
from kfp.deprecated.components import _structures
from kfp.deprecated.containers import entrypoint

V2_COMPONENT_ANNOTATION = 'pipelines.kubeflow.org/component_v2'
_PROGRAM_LAUNCHER_CMD = 'program_path=$(mktemp)\nprintf "%s" "$0" > ' \
                        '"$program_path"\npython3 -u "$program_path" "$@"\n'


class VersionedDependency(object):
    """DependencyVersion specifies the versions."""

    def __init__(self, name, version=None, min_version=None, max_version=None):
        """if version is specified, no need for min_version or max_version; if
        both are specified, version is adopted."""
        self._name = name
        if version is not None:
            self._min_version = version
            self._max_version = version
        else:
            self._min_version = min_version
            self._max_version = max_version

    @property
    def name(self):
        return self._name

    @property
    def min_version(self):
        return self._min_version

    @min_version.setter
    def min_version(self, min_version):
        self._min_version = min_version

    def has_min_version(self):
        return self._min_version != None

    @property
    def max_version(self):
        return self._max_version

    @max_version.setter
    def max_version(self, max_version):
        self._max_version = max_version

    def has_max_version(self):
        return self._max_version != None

    def has_versions(self):
        return (self.has_min_version()) or (self.has_max_version())


class DependencyHelper(object):
    """DependencyHelper manages software dependency information."""

    def __init__(self):
        self._PYTHON_PACKAGE = 'PYTHON_PACKAGE'
        self._dependency = {self._PYTHON_PACKAGE: OrderedDict()}

    @property
    def python_packages(self):
        return self._dependency[self._PYTHON_PACKAGE]

    def add_python_package(self, dependency, override=True):
        """add_single_python_package adds a dependency for the python package.

        Args:
          name: package name
          version: it could be a specific version(1.10.0), or a range(>=1.0,<=2.0)
            if not specified, the default is resolved automatically by the pip system.
          override: whether to override the version if already existing in the dependency.
        """
        if dependency.name in self.python_packages and not override:
            return
        self.python_packages[dependency.name] = dependency

    def generate_pip_requirements(self, target_file):
        """write the python packages to a requirement file the generated file
        follows the order of which the packages are added."""
        with open(target_file, 'w') as f:
            for name, version in self.python_packages.items():
                version = self.python_packages[name]
                version_str = ''
                if version.has_min_version():
                    version_str += ' >= ' + version.min_version + ','
                if version.has_max_version():
                    version_str += ' <= ' + version.max_version + ','
                f.write(name + version_str.rstrip(',') + '\n')


def _dependency_to_requirements(dependency=[], filename='requirements.txt'):
    """
    Generates a requirement file based on the dependency
    Args:
      dependency (list): a list of VersionedDependency, which includes the package name and versions
      filename (str): requirement file name, default as requirements.txt
  """
    dependency_helper = DependencyHelper()
    for version in dependency:
        dependency_helper.add_python_package(version)
    dependency_helper.generate_pip_requirements(filename)


def _generate_dockerfile(filename: str,
                         base_image: str,
                         requirement_filename: Optional[str] = None,
                         add_files: Optional[Dict[str, str]] = None):
    """
    generates dockerfiles
    Args:
      filename (str): target file name for the dockerfile.
      base_image (str): the base image name.
      requirement_filename (str): requirement file name
      add_files (Dict[str, str]): Map containing the files thats should be added to the container. add_files maps the build context relative source paths to the container destination paths.
  """
    with open(filename, 'w') as f:
        f.write('FROM ' + base_image + '\n')
        f.write(
            'RUN apt-get update -y && apt-get install --no-install-recommends -y -q python3 python3-pip python3-setuptools\n'
        )
        if requirement_filename is not None:
            f.write('ADD ' + requirement_filename + ' /ml/requirements.txt\n')
            f.write('RUN python3 -m pip install -r /ml/requirements.txt\n')

        for src_path, dst_path in (add_files or {}).items():
            f.write('ADD ' + src_path + ' ' + dst_path + '\n')


def _configure_logger(logger):
    """_configure_logger configures the logger such that the info level logs go
    to the stdout and the error(or above) level logs go to the stderr.

    It is important for the Jupyter notebook log rendering
    """
    if hasattr(_configure_logger, 'configured'):
        # Skip the logger configuration the second time this function
        # is called to avoid multiple streamhandlers bound to the logger.
        return
    setattr(_configure_logger, 'configured', 'true')
    logger.setLevel(logging.INFO)
    info_handler = logging.StreamHandler(stream=sys.stdout)
    info_handler.addFilter(lambda record: record.levelno <= logging.INFO)
    info_handler.setFormatter(
        logging.Formatter(
            '%(asctime)s:%(levelname)s:%(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'))
    error_handler = logging.StreamHandler(sys.stderr)
    error_handler.addFilter(lambda record: record.levelno > logging.INFO)
    error_handler.setFormatter(
        logging.Formatter(
            '%(asctime)s:%(levelname)s:%(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'))
    logger.addHandler(info_handler)
    logger.addHandler(error_handler)


def _purge_program_launching_code(
        commands: List[str],
        entrypoint_container_path: Optional[str] = None,
        is_v2: bool = False) -> str:
    """Replaces the inline Python code with calling a local program.

    For example,
    Before: sh -ec '... && python3 -u ...' 'import sys ...' --param1 ...
    After:  python -u /ml/main.py --param1 ...

    Args:
      commands: The container commands to be replaced.
      entrypoint_container_path: The path to the entrypoint program in the
        container.
      is_v2: Whether the component being generated is a v2 component. Default is
        False.

    Returns:
      The originally generated inline Python code.
    """
    if not (is_v2 or entrypoint_container_path):
        raise ValueError(
            'Only v2 component has default entrypoint path. '
            'Conventional KFP component needs to specify container '
            'entrypoint explicitly. For example, /ml/main.py')
    program_launcher_index = commands.index(_PROGRAM_LAUNCHER_CMD)
    # When there're preinstallation package specified when converting to component
    # spec the index will be 3, otherwise it'll be 2.
    assert program_launcher_index in [2, 3]
    program_code_index = program_launcher_index + 1
    result = commands[program_code_index]
    if is_v2:
        # TODO: Implement the v2 component entrypoint on KFP.
        # The following are just placeholders.
        commands[program_code_index] = 'kfp.containers.entrypoint'
        commands.pop(program_launcher_index)
        commands[program_launcher_index - 1] = '-m'
        commands[program_launcher_index - 2] = 'python'
    else:
        commands[program_code_index] = entrypoint_container_path
        commands.pop(program_launcher_index)
        commands[program_launcher_index - 1] = '-u'  # -ec => -u
        # sh => python3 or python2
        commands[program_launcher_index - 2] = 'python'

    return result


def build_python_component(
        component_func: Callable,
        target_image: str,
        base_image: Optional[str] = None,
        dependency: Optional[List[VersionedDependency]] = None,
        staging_gcs_path: Optional[str] = None,
        timeout: int = 600,
        namespace: Optional[str] = None,
        target_component_file: Optional[str] = None,
        is_v2: bool = False):
    """build_component automatically builds a container image for the
    component_func based on the base_image and pushes to the target_image.

    Args:
      component_func (python function): The python function to build components
        upon.
      base_image (str): Docker image to use as a base image.
      target_image (str): Full URI to push the target image.
      staging_gcs_path (str): GCS blob that can store temporary build files.
      target_image (str): The target image path.
      timeout (int): The timeout for the image build(in secs), default is 600
        seconds.
      namespace (str): The namespace within which to run the kubernetes Kaniko
        job. If the job is running on GKE and value is None the underlying
        functions will use the default namespace from GKE.
      dependency (list): The list of VersionedDependency, which includes the
        package name and versions, default is empty.
      target_component_file (str): The path to save the generated component YAML
        spec.
      is_v2: Whether or not generating a v2 KFP component, default
        is false.

    Raises:
      ValueError: The function is not decorated with python_component decorator or
        the python_version is neither python2 nor python3
    """

    _configure_logger(logging.getLogger())

    if component_func is None:
        raise ValueError('component_func must not be None')
    if target_image is None:
        raise ValueError('target_image must not be None')

    if staging_gcs_path is None:
        raise ValueError('staging_gcs_path must not be None')

    if base_image is None:
        base_image = getattr(component_func, '_component_base_image', None)
    if base_image is None:
        from ..components._python_op import default_base_image_or_builder
        base_image = default_base_image_or_builder
        if isinstance(base_image, Callable):
            base_image = base_image()
    if not dependency:
        dependency = []

    logging.info('Build an image that is based on ' + base_image +
                 ' and push the image to ' + target_image)

    component_spec = _func_to_component_spec(
        component_func, base_image=base_image)

    if is_v2:
        # TODO: Remove this warning once we make v2 component compatible with KFP
        # v1 stack.
        logging.warning(
            'Currently V2 component is only compatible with v2 KFP.')
        # Annotate the component to be a V2 one.
        if not component_spec.metadata:
            component_spec.metadata = _structures.MetadataSpec()
        if not component_spec.metadata.annotations:
            component_spec.metadata.annotations = {}
        component_spec.metadata.annotations[V2_COMPONENT_ANNOTATION] = 'true'

    command_line_args = component_spec.implementation.container.command

    # The relative path to put the Python program code.
    program_path = 'ml/main.py'
    # The relative path used when building a V2 component.
    v2_entrypoint_path = None
    # Python program code extracted from the component spec.
    program_code = None

    if is_v2:

        program_code = _purge_program_launching_code(
            commands=command_line_args, is_v2=True)

        # Override user program args for new-styled component.
        # TODO: The actual program args will be changed after we support v2
        # component on KFP.
        # For v2 component, the received command line args are fixed as follows:
        # --executor_input_str
        # {Executor input pb message at runtime}
        # --function_name
        # {The name of user defined function}
        # --output_metadata_path
        # {The place to write output metadata JSON file}
        program_args = [
            '--executor_input_str',
            _structures.ExecutorInputPlaceholder(),
            '--{}'.format(entrypoint.FN_NAME_ARG), component_func.__name__,
            '--output_metadata_path',
            _structures.OutputMetadataPlaceholder()
        ]

        component_spec.implementation.container.args = program_args
    else:
        program_code = _purge_program_launching_code(
            commands=command_line_args,
            entrypoint_container_path='/' + program_path)

    arc_docker_filename = 'Dockerfile'
    arc_requirement_filename = 'requirements.txt'

    with tempfile.TemporaryDirectory() as local_build_dir:
        # Write the program code to a file in the context directory
        local_python_filepath = os.path.join(local_build_dir, program_path)
        os.makedirs(os.path.dirname(local_python_filepath), exist_ok=True)

        with open(local_python_filepath, 'w') as f:
            f.write(program_code)

        # Generate the python package requirements file in the context directory
        local_requirement_filepath = os.path.join(local_build_dir,
                                                  arc_requirement_filename)
        if is_v2:
            # For v2 components, KFP are expected to be packed in the container.
            dependency.append(
                VersionedDependency(name='kfp', min_version='1.4.0'))

        _dependency_to_requirements(dependency, local_requirement_filepath)

        # Generate Dockerfile in the context directory
        local_docker_filepath = os.path.join(local_build_dir,
                                             arc_docker_filename)
        add_files = {program_path: '/' + program_path}

        _generate_dockerfile(
            local_docker_filepath,
            base_image,
            arc_requirement_filename,
            add_files=add_files)

        logging.info('Building and pushing container image.')
        container_builder = ContainerBuilder(staging_gcs_path, target_image,
                                             namespace)
        image_name_with_digest = container_builder.build(
            local_build_dir, arc_docker_filename, target_image, timeout)

    component_spec.implementation.container.image = image_name_with_digest

    # Optionally writing the component definition to a local file for sharing
    target_component_file = target_component_file or getattr(
        component_func, '_component_target_component_file', None)
    if target_component_file:
        component_spec.save(target_component_file)

    task_factory_function = _create_task_factory_from_component_spec(
        component_spec)
    return task_factory_function


@deprecated(
    version='0.1.32',
    reason='`build_docker_image` is deprecated. Use `kfp.containers.build_image_from_working_dir` instead.'
)
def build_docker_image(staging_gcs_path,
                       target_image,
                       dockerfile_path,
                       timeout=600,
                       namespace=None):
    """build_docker_image automatically builds a container image based on the
    specification in the dockerfile and pushes to the target_image.

    Args:
      staging_gcs_path (str): GCS blob that can store temporary build files
      target_image (str): gcr path to push the final image
      dockerfile_path (str): local path to the dockerfile
      timeout (int): the timeout for the image build(in secs), default is 600 seconds
      namespace (str): the namespace within which to run the kubernetes kaniko job. Default is None. If the
      job is running on GKE and value is None the underlying functions will use the default namespace from GKE.
    """
    _configure_logger(logging.getLogger())

    with tempfile.TemporaryDirectory() as local_build_dir:
        dockerfile_rel_path = 'Dockerfile'
        dst_dockerfile_path = os.path.join(local_build_dir, dockerfile_rel_path)
        shutil.copyfile(dockerfile_path, dst_dockerfile_path)

        container_builder = ContainerBuilder(
            staging_gcs_path, target_image, namespace=namespace)
        image_name_with_digest = container_builder.build(
            local_build_dir, dockerfile_rel_path, target_image, timeout)

    logging.info('Build image complete.')
    return image_name_with_digest



# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.structures."""

import os
import tempfile
import textwrap
import unittest

from absl.testing import parameterized
from kfp import compiler
from kfp import components
from kfp import dsl
from kfp.components import component_factory
from kfp.components import placeholders
from kfp.components import structures

V1_YAML_IF_PLACEHOLDER = textwrap.dedent("""\
    implementation:
      container:
        args:
        - if:
            cond:
              isPresent: optional_input_1
            else:
              - --arg2
              - default
            then:
              - --arg1
              - {inputUri: optional_input_1}
        image: alpine
    inputs:
    - {name: optional_input_1, optional: true, type: String}
    name: component_if
    """)

COMPONENT_SPEC_IF_PLACEHOLDER = structures.ComponentSpec(
    name='component_if',
    implementation=structures.Implementation(
        container=structures.ContainerSpecImplementation(
            image='alpine',
            args=[
                placeholders.IfPresentPlaceholder(
                    input_name='optional_input_1',
                    then=[
                        '--arg1',
                        placeholders.InputUriPlaceholder(
                            input_name='optional_input_1'),
                    ],
                    else_=[
                        '--arg2',
                        'default',
                    ])
            ])),
    inputs={
        'optional_input_1': structures.InputSpec(type='String', default=None)
    },
)

V1_YAML_CONCAT_PLACEHOLDER = textwrap.dedent("""\
    name: component_concat
    implementation:
      container:
        args:
        - concat: ['--arg1', {inputValue: input_prefix}]
        image: alpine
    inputs:
    - {name: input_prefix, type: String}
    """)

COMPONENT_SPEC_CONCAT_PLACEHOLDER = structures.ComponentSpec(
    name='component_concat',
    implementation=structures.Implementation(
        container=structures.ContainerSpecImplementation(
            image='alpine',
            args=[
                placeholders.ConcatPlaceholder(items=[
                    '--arg1',
                    placeholders.InputValuePlaceholder(
                        input_name='input_prefix'),
                ])
            ])),
    inputs={'input_prefix': structures.InputSpec(type='String')},
)

V1_YAML_NESTED_PLACEHOLDER = textwrap.dedent("""\
    name: component_nested
    implementation:
      container:
        args:
        - concat:
            - --arg1
            - if:
                cond:
                    isPresent: input_prefix
                then: {inputValue: input_prefix}
                else: default
        image: alpine
    inputs:
    - {name: input_prefix, optional: false, type: String}
    """)

COMPONENT_SPEC_NESTED_PLACEHOLDER = structures.ComponentSpec(
    name='component_nested',
    implementation=structures.Implementation(
        container=structures.ContainerSpecImplementation(
            image='alpine',
            args=[
                placeholders.ConcatPlaceholder(items=[
                    '--arg1',
                    placeholders.IfPresentPlaceholder(
                        input_name='input_prefix',
                        then=placeholders.InputValuePlaceholder(
                            input_name='input_prefix'),
                        else_='default'),
                ]),
            ])),
    inputs={'input_prefix': structures.InputSpec(type='String')},
)

V1_YAML_EXECUTOR_INPUT_PLACEHOLDER = textwrap.dedent("""\
    name: component_executor_input
    inputs:
    - {name: input, type: String}
    implementation:
      container:
        image: alpine
        command:
        - python
        - -m
        - kfp.containers.entrypoint
        args:
        - --executor_input
        - {executorInput: null}
        - --function_name
        - test_function
    """)

COMPONENT_SPEC_EXECUTOR_INPUT_PLACEHOLDER = structures.ComponentSpec(
    name='component_executor_input',
    implementation=structures.Implementation(
        container=structures.ContainerSpecImplementation(
            image='alpine',
            command=[
                'python',
                '-m',
                'kfp.containers.entrypoint',
            ],
            args=[
                '--executor_input',
                placeholders.ExecutorInputPlaceholder(),
                '--function_name',
                'test_function',
            ])),
    inputs={'input': structures.InputSpec(type='String')},
)

V1_NONCANONICAL_GENERIC_TYPES_COMPONENT_SPEC = textwrap.dedent("""\
    name: generic_types_test
    inputs:
    - {name: input1, type: "List[str]"}
    - {name: input2, type: "typing.List[str]"}
    - {name: input3, type: "Dict[str, str]"}
    - {name: input4, type: "typing.Dict[str, str]"}
    outputs:
    - {name: output1, type: "List[str]"}
    - {name: output2, type: "typing.List[str]"}
    - {name: output3, type: "Dict[str, str]"}
    - {name: output4, type: "typing.Dict[str, str]"}
    implementation:
      container:
        image: alpine
    """)


class StructuresTest(parameterized.TestCase):

    def test_component_spec_with_placeholder_referencing_nonexisting_input_output(
            self):
        with self.assertRaisesRegex(
                ValueError,
                r'^Argument "InputValuePlaceholder" references nonexistant input: "input000".'
        ):
            structures.ComponentSpec(
                name='component_1',
                implementation=structures.Implementation(
                    container=structures.ContainerSpecImplementation(
                        image='alpine',
                        command=[
                            'sh',
                            '-c',
                            'set -ex\necho "$0" > "$1"',
                            placeholders.InputValuePlaceholder(
                                input_name='input000'),
                            placeholders.OutputPathPlaceholder(
                                output_name='output1'),
                        ],
                    )),
                inputs={'input1': structures.InputSpec(type='String')},
                outputs={'output1': structures.OutputSpec(type='String')},
            )

        with self.assertRaisesRegex(
                ValueError,
                r'^Argument "OutputPathPlaceholder" references nonexistant output: "output000".'
        ):
            structures.ComponentSpec(
                name='component_1',
                implementation=structures.Implementation(
                    container=structures.ContainerSpecImplementation(
                        image='alpine',
                        command=[
                            'sh',
                            '-c',
                            'set -ex\necho "$0" > "$1"',
                            placeholders.InputValuePlaceholder(
                                input_name='input1'),
                            placeholders.OutputPathPlaceholder(
                                output_name='output000'),
                        ],
                    )),
                inputs={'input1': structures.InputSpec(type='String')},
                outputs={'output1': structures.OutputSpec(type='String')},
            )

    def test_simple_component_spec_save_to_component_yaml(self):
        # tests writing old style (less verbose) and reading in new style (more verbose)
        original_component_spec = structures.ComponentSpec(
            name='component_1',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=[
                        'sh', '-c', 'set -ex\necho "$0" > "$1"',
                        "{{$.inputs.parameters['input1']}}",
                        "{{$.outputs.parameters['output1'].output_file}}"
                    ],
                )),
            inputs={'input1': structures.InputSpec(type='String')},
            outputs={'output1': structures.OutputSpec(type='String')},
        )

        test_component = components.PythonComponent(
            # dummy python_func not used in behavior that is being tested
            python_func=lambda: None,
            component_spec=original_component_spec,
        )
        with tempfile.TemporaryDirectory() as tempdir:
            output_path = os.path.join(tempdir, 'component.yaml')
            compiler.Compiler().compile(test_component, output_path)

            # test that it can be read back correctly
            with open(output_path, 'r') as f:
                contents = f.read()
            new_component_spec = structures.ComponentSpec.from_yaml_documents(
                contents)

        self.assertEqual(original_component_spec, new_component_spec)

    def test_simple_component_spec_load_from_v2_component_yaml(self):
        component_yaml_v2 = textwrap.dedent("""\
components:
  comp-component-1:
    executorLabel: exec-component-1
    inputDefinitions:
      parameters:
        input1:
          parameterType: STRING
    outputDefinitions:
      parameters:
        output1:
          parameterType: STRING
deploymentSpec:
  executors:
    exec-component-1:
      container:
        command:
        - sh
        - -c
        - 'set -ex

          echo "$0" > "$1"'
        - '{{$.inputs.parameters[''input1'']}}'
        - '{{$.outputs.parameters[''output1''].output_file}}'
        image: alpine
pipelineInfo:
  name: component-1
root:
  dag:
    tasks:
      component-1:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-component-1
        inputs:
          parameters:
            input1:
              componentInputParameter: input1
        taskInfo:
          name: component-1
  inputDefinitions:
    parameters:
      input1:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2
        """)

        generated_spec = structures.ComponentSpec.from_yaml_documents(
            component_yaml_v2)

        expected_spec = structures.ComponentSpec(
            name='component-1',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=[
                        'sh',
                        '-c',
                        'set -ex\necho "$0" > "$1"',
                        "{{$.inputs.parameters['input1']}}",
                        "{{$.outputs.parameters['output1'].output_file}}",
                    ],
                )),
            inputs={'input1': structures.InputSpec(type='String')},
            outputs={'output1': structures.OutputSpec(type='String')})

        self.assertEqual(generated_spec, expected_spec)

    @parameterized.parameters(
        {
            'yaml': V1_YAML_IF_PLACEHOLDER,
            'expected_component': COMPONENT_SPEC_IF_PLACEHOLDER
        },
        {
            'yaml': V1_YAML_CONCAT_PLACEHOLDER,
            'expected_component': COMPONENT_SPEC_CONCAT_PLACEHOLDER
        },
        {
            'yaml': V1_YAML_NESTED_PLACEHOLDER,
            'expected_component': COMPONENT_SPEC_NESTED_PLACEHOLDER
        },
        {
            'yaml': V1_YAML_EXECUTOR_INPUT_PLACEHOLDER,
            'expected_component': COMPONENT_SPEC_EXECUTOR_INPUT_PLACEHOLDER
        },
    )
    def test_component_spec_placeholder_load_from_v2_component_yaml(
            self, yaml, expected_component):
        generated_spec = structures.ComponentSpec.from_yaml_documents(yaml)
        self.assertEqual(generated_spec, expected_component)

    def test_component_spec_load_from_v1_component_yaml(self):
        component_yaml_v1 = textwrap.dedent("""\
        name: Component with 2 inputs and 2 outputs
        inputs:
        - {name: Input parameter, type: String}
        - {name: Input artifact}
        outputs:
        - {name: Output 1}
        - {name: Output 2}
        implementation:
          container:
            image: busybox
            command: [sh, -c, '
                mkdir -p $(dirname "$2")
                mkdir -p $(dirname "$3")
                echo "$0" > "$2"
                cp "$1" "$3"
                '
            ]
            args:
            - {inputValue: Input parameter}
            - {inputPath: Input artifact}
            - {outputPath: Output 1}
            - {outputPath: Output 2}
        """)

        generated_spec = structures.ComponentSpec.from_yaml_documents(
            component_yaml_v1)

        expected_spec = structures.ComponentSpec(
            name='Component with 2 inputs and 2 outputs',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='busybox',
                    command=[
                        'sh',
                        '-c',
                        (' mkdir -p $(dirname "$2") mkdir -p $(dirname "$3") '
                         'echo "$0" > "$2" cp "$1" "$3" '),
                    ],
                    args=[
                        placeholders.InputValuePlaceholder(
                            input_name='input_parameter'),
                        placeholders.InputPathPlaceholder(
                            input_name='input_artifact'),
                        placeholders.OutputPathPlaceholder(
                            output_name='output_1'),
                        placeholders.OutputPathPlaceholder(
                            output_name='output_2'),
                    ],
                    env={},
                )),
            inputs={
                'input_parameter':
                    structures.InputSpec(type='String'),
                'input_artifact':
                    structures.InputSpec(type='system.Artifact@0.0.1')
            },
            outputs={
                'output_1': structures.OutputSpec(type='system.Artifact@0.0.1'),
                'output_2': structures.OutputSpec(type='system.Artifact@0.0.1'),
            })
        self.assertEqual(generated_spec, expected_spec)


class TestContainerSpecImplementation(unittest.TestCase):

    def test_command_and_args(self):
        obj = structures.ContainerSpecImplementation(
            image='image', command=['command'], args=['args'])
        self.assertEqual(obj.command, ['command'])
        self.assertEqual(obj.args, ['args'])

        obj = structures.ContainerSpecImplementation(
            image='image', command=[], args=[])
        self.assertEqual(obj.command, None)
        self.assertEqual(obj.args, None)

    def test_env(self):
        obj = structures.ContainerSpecImplementation(
            image='image',
            command=['command'],
            args=['args'],
            env={'env': 'env'})
        self.assertEqual(obj.env, {'env': 'env'})

        obj = structures.ContainerSpecImplementation(
            image='image', command=[], args=[], env={})
        self.assertEqual(obj.env, None)

    def test_from_container_dict_no_placeholders(self):
        expected_container_spec = structures.ContainerSpecImplementation(
            image='python:3.7',
            command=['sh', '-c', 'dummy'],
            args=['--executor_input', '{{$}}', '--function_to_execute', 'func'],
            env={'ENV1': 'val1'},
            resources=None)

        container_dict = {
            'args': [
                '--executor_input', '{{$}}', '--function_to_execute', 'func'
            ],
            'command': ['sh', '-c', 'dummy'],
            'image': 'python:3.7',
            'env': {
                'ENV1': 'val1'
            },
        }

        loaded_container_spec = structures.ContainerSpecImplementation.from_container_dict(
            container_dict)
        self.assertEqual(expected_container_spec, loaded_container_spec)

    def test_raise_error_if_access_artifact_by_itself(self):

        def comp_with_artifact_input(dataset: dsl.Input[dsl.Dataset]):
            return dsl.ContainerSpec(
                image='gcr.io/my-image',
                command=['sh', 'run.sh'],
                args=[dataset])

        def comp_with_artifact_output(dataset_old: dsl.Output[dsl.Dataset],
                                      dataset_new: dsl.Output[dsl.Dataset],
                                      optional_input: str = 'default'):
            return dsl.ContainerSpec(
                image='gcr.io/my-image',
                command=['sh', 'run.sh'],
                args=[
                    dsl.IfPresentPlaceholder(
                        input_name='optional_input',
                        then=[dataset_old],
                        else_=[dataset_new])
                ])

        self.assertRaisesRegex(
            ValueError,
            r'Cannot access artifact by itself in the container definition.',
            component_factory.create_container_component_from_func,
            comp_with_artifact_input)
        self.assertRaisesRegex(
            ValueError,
            r'Cannot access artifact by itself in the container definition.',
            component_factory.create_container_component_from_func,
            comp_with_artifact_output)


class TestComponentSpec(unittest.TestCase):

    def test_inputs(self):
        obj = structures.ComponentSpec(
            name='name',
            implementation=structures.Implementation(container=None),
            inputs={})
        self.assertEqual(obj.inputs, None)

    def test_outputs(self):
        obj = structures.ComponentSpec(
            name='name',
            implementation=structures.Implementation(container=None),
            outputs={})
        self.assertEqual(obj.outputs, None)


class TestInputSpec(unittest.TestCase):

    def test_equality(self):
        self.assertEqual(
            structures.InputSpec(type='String', default=None),
            structures.InputSpec(type='String', default=None))
        self.assertNotEqual(
            structures.InputSpec(type='String', default=None),
            structures.InputSpec(type='String', default='test', optional=True))
        self.assertEqual(
            structures.InputSpec(type='List', default=None),
            structures.InputSpec(type='typing.List', default=None))
        self.assertEqual(
            structures.InputSpec(type='List', default=None),
            structures.InputSpec(type='typing.List[int]', default=None))
        self.assertEqual(
            structures.InputSpec(type='List'),
            structures.InputSpec(type='typing.List[typing.Dict[str, str]]'))

    def test_optional(self):
        input_spec = structures.InputSpec(
            type='String', default=None, optional=True)
        self.assertEqual(input_spec.default, None)
        self.assertEqual(input_spec.optional, True)

        input_spec = structures.InputSpec(type='String')
        self.assertEqual(input_spec.default, None)
        self.assertEqual(input_spec.optional, False)

    def test_from_ir_component_inputs_dict(self):
        parameter_dict = {'parameterType': 'STRING'}
        input_spec = structures.InputSpec.from_ir_component_inputs_dict(
            parameter_dict)
        self.assertEqual(input_spec.type, 'String')
        self.assertEqual(input_spec.default, None)

        parameter_dict = {'parameterType': 'NUMBER_INTEGER'}
        input_spec = structures.InputSpec.from_ir_component_inputs_dict(
            parameter_dict)
        self.assertEqual(input_spec.type, 'Integer')
        self.assertEqual(input_spec.default, None)

        parameter_dict = {
            'defaultValue': 'default value',
            'parameterType': 'STRING'
        }
        input_spec = structures.InputSpec.from_ir_component_inputs_dict(
            parameter_dict)
        self.assertEqual(input_spec.type, 'String')
        self.assertEqual(input_spec.default, 'default value')

        input_spec = structures.InputSpec.from_ir_component_inputs_dict(
            parameter_dict)
        self.assertEqual(input_spec.type, 'String')
        self.assertEqual(input_spec.default, 'default value')

        artifact_dict = {
            'artifactType': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            }
        }
        input_spec = structures.InputSpec.from_ir_component_inputs_dict(
            artifact_dict)
        self.assertEqual(input_spec.type, 'system.Artifact@0.0.1')

    def test_assert_optional_must_be_true_when_default_is_not_none(self):
        with self.assertRaisesRegex(ValueError,
                                    r'must be True if `default` is not None'):
            input_spec = structures.InputSpec(type='String', default='text')


class TestOutputSpec(parameterized.TestCase):

    def test_from_ir_component_outputs_dict(self):
        parameter_dict = {'parameterType': 'STRING'}
        output_spec = structures.OutputSpec.from_ir_component_outputs_dict(
            parameter_dict)
        self.assertEqual(output_spec.type, 'String')

        artifact_dict = {
            'artifactType': {
                'schemaTitle': 'system.Artifact',
                'schemaVersion': '0.0.1'
            }
        }
        output_spec = structures.OutputSpec.from_ir_component_outputs_dict(
            artifact_dict)
        self.assertEqual(output_spec.type, 'system.Artifact@0.0.1')


V1_YAML = textwrap.dedent("""\
    implementation:
      container:
        args:
        - if:
            cond:
              isPresent: optional_input_1
            else:
              - --arg2
              - default
            then:
              - --arg1
              - {inputUri: optional_input_1}
        image: alpine
    inputs:
    - {name: optional_input_1, optional: true, type: String}
    name: component_if
    """)


class TestReadInComponent(parameterized.TestCase):

    def test_read_v1(self):
        component_spec = structures.ComponentSpec.from_yaml_documents(
            V1_YAML_IF_PLACEHOLDER)
        self.assertEqual(component_spec.name, 'component-if')
        self.assertEqual(component_spec.implementation.container.image,
                         'alpine')

    def test_simple_placeholder(self):
        compiled_yaml = textwrap.dedent("""
components:
  comp-component1:
    executorLabel: exec-component1
    inputDefinitions:
      parameters:
        input1:
          parameterType: STRING
    outputDefinitions:
      artifacts:
        output1:
          artifactType:
            schemaTitle: system.Artifact
            schemaVersion: 0.0.1
deploymentSpec:
  executors:
    exec-component1:
      container:
        args:
        - '{{$.inputs.parameters[''input1'']}}'
        - '{{$.outputs.artifacts[''output1''].path}}'
        command:
        - sh
        - -c
        - echo "$0" >> "$1"
        image: alpine
pipelineInfo:
  name: component1
root:
  dag:
    tasks:
      component1:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-component1
        inputs:
          parameters:
            input1:
              componentInputParameter: input1
        taskInfo:
          name: component1
  inputDefinitions:
    parameters:
      input1:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2""")
        loaded_component_spec = structures.ComponentSpec.from_yaml_documents(
            compiled_yaml)
        component_spec = structures.ComponentSpec(
            name='component1',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=['sh', '-c', 'echo "$0" >> "$1"'],
                    args=[
                        "{{$.inputs.parameters['input1']}}",
                        "{{$.outputs.artifacts['output1'].path}}",
                    ],
                    env=None,
                    resources=None),
                graph=None,
                importer=None),
            description=None,
            inputs={
                'input1': structures.InputSpec(type='String', default=None)
            },
            outputs={
                'output1': structures.OutputSpec(type='system.Artifact@0.0.1')
            })
        self.assertEqual(loaded_component_spec, component_spec)

    def test_if_placeholder(self):
        compiled_yaml = textwrap.dedent("""
components:
  comp-if:
    executorLabel: exec-if
    inputDefinitions:
      parameters:
        optional_input_1:
          parameterType: STRING
deploymentSpec:
  executors:
    exec-if:
      container:
        args:
        - 'input: '
        - '{{$.inputs.parameters[''optional_input_1'']}}'
        command:
        - sh
        - -c
        - echo "$0" "$1"
        image: alpine
pipelineInfo:
  name: if
root:
  dag:
    tasks:
      if:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-if
        inputs:
          parameters:
            optional_input_1:
              componentInputParameter: optional_input_1
        taskInfo:
          name: if
  inputDefinitions:
    parameters:
      optional_input_1:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2""")
        loaded_component_spec = structures.ComponentSpec.from_yaml_documents(
            compiled_yaml)
        component_spec = structures.ComponentSpec(
            name='if',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=['sh', '-c', 'echo "$0" "$1"'],
                    args=[
                        'input: ',
                        "{{$.inputs.parameters['optional_input_1']}}",
                    ],
                    env=None,
                    resources=None),
                graph=None,
                importer=None),
            description=None,
            inputs={
                'optional_input_1':
                    structures.InputSpec(type='String', default=None)
            },
            outputs=None)
        self.assertEqual(loaded_component_spec, component_spec)

    def test_concat_placeholder(self):
        compiled_yaml = textwrap.dedent("""
components:
  comp-concat:
    executorLabel: exec-concat
    inputDefinitions:
      parameters:
        input1:
          parameterType: STRING
        input2:
          parameterType: STRING
deploymentSpec:
  executors:
    exec-concat:
      container:
        command:
        - sh
        - -c
        - echo "$0"
        - '{{$.inputs.parameters[''input1'']}}+{{$.inputs.parameters[''input2'']}}'
        image: alpine
pipelineInfo:
  name: concat
root:
  dag:
    tasks:
      concat:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-concat
        inputs:
          parameters:
            input1:
              componentInputParameter: input1
            input2:
              componentInputParameter: input2
        taskInfo:
          name: concat
  inputDefinitions:
    parameters:
      input1:
        parameterType: STRING
      input2:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2""")
        loaded_component_spec = structures.ComponentSpec.from_yaml_documents(
            compiled_yaml)
        component_spec = structures.ComponentSpec(
            name='concat',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=[
                        'sh',
                        '-c',
                        'echo "$0"',
                        "{{$.inputs.parameters['input1']}}+{{$.inputs.parameters['input2']}}",
                    ],
                    args=None,
                    env=None,
                    resources=None),
                graph=None,
                importer=None),
            description=None,
            inputs={
                'input1': structures.InputSpec(type='String', default=None),
                'input2': structures.InputSpec(type='String', default=None)
            },
            outputs=None)
        self.assertEqual(loaded_component_spec, component_spec)


class TestNormalizeTimeString(parameterized.TestCase):

    @parameterized.parameters([
        ('1 hour', '1h'),
        ('2 hours', '2h'),
        ('2hours', '2h'),
        ('2 w', '2w'),
        ('2d', '2d'),
    ])
    def test(self, unnorm: str, norm: str):
        self.assertEqual(structures.normalize_time_string(unnorm), norm)

    def test_multipart_duration_raises(self):
        with self.assertRaisesRegex(ValueError, 'Invalid duration string:'):
            structures.convert_duration_to_seconds('1 day 1 hour')

    def test_non_int_value_raises(self):
        with self.assertRaisesRegex(ValueError, 'Invalid duration string:'):
            structures.convert_duration_to_seconds('one hour')


class TestConvertDurationToSeconds(parameterized.TestCase):

    @parameterized.parameters([
        ('1 hour', 3600),
        ('2 hours', 7200),
        ('2hours', 7200),
        ('2 w', 1209600),
        ('2d', 172800),
    ])
    def test(self, duration: str, seconds: int):
        self.assertEqual(
            structures.convert_duration_to_seconds(duration), seconds)

    def test_unsupported_duration_unit(self):
        with self.assertRaisesRegex(ValueError, 'Unsupported duration unit:'):
            structures.convert_duration_to_seconds('1 year')


class TestRetryPolicy(unittest.TestCase):

    def test_to_proto(self):
        retry_policy_struct = structures.RetryPolicy(
            max_retry_count=10,
            backoff_duration='1h',
            backoff_factor=1.5,
            backoff_max_duration='2 weeks')

        retry_policy_proto = retry_policy_struct.to_proto()
        self.assertEqual(retry_policy_proto.max_retry_count, 10)
        self.assertEqual(retry_policy_proto.backoff_duration.seconds, 3600)
        self.assertEqual(retry_policy_proto.backoff_factor, 1.5)
        # tests cap
        self.assertEqual(retry_policy_proto.backoff_max_duration.seconds, 3600)


class TestDeserializeV1ComponentYamlDefaults(unittest.TestCase):

    def test_True(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Boolean, default: "True" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, True)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.bool_value, True)

    def test_true(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Boolean, default: "true" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, True)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.bool_value, True)

    def test_false(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Boolean, default: "false" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, False)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.bool_value, False)

    def test_False(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Boolean, default: "False" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, False)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.bool_value, False)

    def test_int(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Integer, default: "1" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, 1)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.number_value, 1.0)

    def test_float(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: Float, default: "1.0" }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, 1.0)
        self.assertEqual(
            comp.pipeline_spec.root.input_definitions.parameters['val']
            .default_value.number_value, 1.0)

    def test_struct(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: JsonObject, default: '{"a": 1.0}' }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, {'a': 1.0})
        self.assertEqual(
            dict(comp.pipeline_spec.root.input_definitions.parameters['val']
                 .default_value.struct_value), {'a': 1.0})

    def test_array(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: JsonObject, default: '["a", 1.0]' }
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputValue: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertEqual(comp.component_spec.inputs['val'].default, ['a', 1.0])
        self.assertEqual(
            list(comp.pipeline_spec.root.input_definitions.parameters['val']
                 .default_value.list_value), ['a', 1.0])

    def test_artifact_with_dict_type_passes_through(self):
        comp_text = textwrap.dedent("""\
inputs:
  - { name: val, type: {Key: Val}}
implementation:
  container:
    image: alpine
    command:
      - sh
      - -c
      - |
        echo $0
      - { inputPath: val }
""")
        comp = components.load_component_from_text(comp_text)
        self.assertFalse(comp.component_spec.inputs['val'].optional)
        self.assertFalse(comp.pipeline_spec.root.input_definitions
                         .artifacts['val'].is_optional)

    def test_load_noncanonical_v1_generic_types(self):
        loaded_comp = components.load_component_from_text(
            V1_NONCANONICAL_GENERIC_TYPES_COMPONENT_SPEC)
        inputs = loaded_comp.component_spec.inputs
        outputs = loaded_comp.component_spec.outputs
        self.assertEqual(inputs['input1'].type, 'List')
        self.assertEqual(inputs['input2'].type, 'List')
        self.assertEqual(inputs['input3'].type, 'Dict')
        self.assertEqual(inputs['input4'].type, 'Dict')

        self.assertEqual(outputs['output1'].type, 'List')
        self.assertEqual(outputs['output2'].type, 'List')
        self.assertEqual(outputs['output3'].type, 'Dict')
        self.assertEqual(outputs['output4'].type, 'Dict')


class TestLoadDocumentsFromYAML(unittest.TestCase):

    def test_no_documents(self):
        with self.assertRaisesRegex(
                ValueError,
                r'Expected one or two YAML documents in the IR YAML file\. Got\: 0\.'
        ):
            structures.load_documents_from_yaml('')

    def test_one_document(self):
        doc1, doc2 = structures.load_documents_from_yaml(
            textwrap.dedent("""\
            key1: value1
            """))
        self.assertEqual(doc1, {'key1': 'value1'})
        self.assertEqual(doc2, {})

    def test_two_documents(self):
        doc1, doc2 = structures.load_documents_from_yaml(
            textwrap.dedent("""\
            key1: value1
            ---
            key2: value2
            """))
        self.assertEqual(doc1, {'key1': 'value1'})
        self.assertEqual(doc2, {'key2': 'value2'})

    def test_three_documents(self):
        with self.assertRaisesRegex(
                ValueError,
                r'Expected one or two YAML documents in the IR YAML file\. Got\: 3\.'
        ):
            structures.load_documents_from_yaml(
                textwrap.dedent("""\
                key3: value3
                ---
                key3: value3
                ---
                key3: value3
                """))


if __name__ == '__main__':
    unittest.main()



"""CLI interface for KFP diagnose_me tool."""

import json as json_library
import sys
from typing import Dict, List, Text, Union

import click
from kfp.cli.diagnose_me import dev_env
from kfp.cli.diagnose_me import gcp
from kfp.cli.diagnose_me import kubernetes_cluster
from kfp.cli.diagnose_me import kubernetes_cluster as k8
from kfp.cli.diagnose_me import utility

ResultsType = Dict[Union[gcp.Commands, dev_env.Commands,
                         kubernetes_cluster.Commands], utility.ExecutorResponse]


@click.command()
@click.option(
    '-j',
    '--json',
    is_flag=True,
    help='Output in Json format, human readable format is set by default.')
@click.option(
    '-p',
    '--project-id',
    type=Text,
    help='Target project id. It will use environment default if not specified.')
@click.option(
    '-n',
    '--namespace',
    type=Text,
    help='Namespace to use for Kubernetes cluster.all-namespaces is used if not specified.'
)
@click.pass_context
def diagnose_me(ctx: click.Context, json: bool, project_id: str,
                namespace: str):
    """Runs KFP environment diagnostic."""
    # validate kubectl, gcloud , and gsutil exist
    local_env_gcloud_sdk = gcp.get_gcp_configuration(
        gcp.Commands.GET_GCLOUD_VERSION,
        project_id=project_id,
        human_readable=False)
    for app in ['Google Cloud SDK', 'gsutil', 'kubectl']:
        if app not in local_env_gcloud_sdk.json_output:
            raise RuntimeError(
                f'{app} is not installed, gcloud, gsutil and kubectl are required '
                + 'for this app to run. Please follow instructions at ' +
                'https://cloud.google.com/sdk/install to install the SDK.')

    click.echo('Collecting diagnostic information ...', file=sys.stderr)

    # default behaviour dump all configurations
    results: ResultsType = {
        gcp_command: gcp.get_gcp_configuration(
            gcp_command, project_id=project_id, human_readable=not json)
        for gcp_command in gcp.Commands
    }

    for k8_command in k8.Commands:
        results[k8_command] = k8.get_kubectl_configuration(
            k8_command, human_readable=not json)

    for dev_env_command in dev_env.Commands:
        results[dev_env_command] = dev_env.get_dev_env_configuration(
            dev_env_command, human_readable=not json)

    print_to_sdtout(results, not json)


def print_to_sdtout(results: ResultsType, human_readable: bool):
    """Viewer to print the ExecutorResponse results to stdout.

    Args:
      results: A dictionary with key:command names and val: Execution response
      human_readable: Print results in human readable format. If set to True
        command names will be printed as visual delimiters in new lines. If False
        results are printed as a dictionary with command as key.
    """

    output_dict = {}
    human_readable_result: List[str] = []
    for key, val in results.items():
        if val.has_error:
            output_dict[
                key.
                name] = f'Following error occurred during the diagnoses: {val.stderr}'
            continue

        output_dict[key.name] = val.json_output
        human_readable_result.extend(
            (f'================ {key.name} ===================',
             val.parsed_output))

    if human_readable:
        result = '\n'.join(human_readable_result)
    else:
        result = json_library.dumps(
            output_dict, sort_keys=True, indent=2, separators=(',', ': '))

    click.echo(result)



import click
import json
from typing import List

from kfp.deprecated.cli.output import print_output, OutputFormat
import kfp_server_api
from kfp_server_api.models.api_experiment import ApiExperiment


@click.group()
def experiment():
    """Manage experiment resources."""


@experiment.command()
@click.option('-d', '--description', help="Description of the experiment.")
@click.argument("name")
@click.pass_context
def create(ctx: click.Context, description: str, name: str):
    """Create an experiment."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.create_experiment(name, description=description)
    _display_experiment(response, output_format)


@experiment.command()
@click.option(
    '--page-token', default='', help="Token for starting of the page.")
@click.option(
    '-m', '--max-size', default=100, help="Max size of the listed experiments.")
@click.option(
    '--sort-by',
    default="created_at desc",
    help="Can be '[field_name]', '[field_name] desc'. For example, 'name desc'."
)
@click.option(
    '--filter',
    help=(
        "filter: A url-encoded, JSON-serialized Filter protocol buffer "
        "(see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto))."
    ))
@click.pass_context
def list(ctx: click.Context, page_token: str, max_size: int, sort_by: str,
         filter: str):
    """List experiments."""
    client = ctx.obj['client']
    output_format = ctx.obj['output']

    response = client.list_experiments(
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    if response.experiments:
        _display_experiments(response.experiments, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = "No experiments found"
        click.echo(msg)


@experiment.command()
@click.argument("experiment-id")
@click.pass_context
def get(ctx: click.Context, experiment_id: str):
    """Get detailed information about an experiment."""
    client = ctx.obj["client"]
    output_format = ctx.obj["output"]

    response = client.get_experiment(experiment_id)
    _display_experiment(response, output_format)


@experiment.command()
@click.argument("experiment-id")
@click.pass_context
def delete(ctx: click.Context, experiment_id: str):
    """Delete an experiment."""

    confirmation = "Caution. The RunDetails page could have an issue" \
                   " when it renders a run that has no experiment." \
                   " Do you want to continue?"
    if not click.confirm(confirmation):
        return

    client = ctx.obj["client"]

    client.delete_experiment(experiment_id)
    click.echo("{} is deleted.".format(experiment_id))


def _display_experiments(experiments: List[ApiExperiment],
                         output_format: OutputFormat):
    headers = ["Experiment ID", "Name", "Created at"]
    data = [
        [exp.id, exp.name, exp.created_at.isoformat()] for exp in experiments
    ]
    print_output(data, headers, output_format, table_format="grid")


def _display_experiment(exp: kfp_server_api.ApiExperiment,
                        output_format: OutputFormat):
    table = [
        ["ID", exp.id],
        ["Name", exp.name],
        ["Description", exp.description],
        ["Created at", exp.created_at.isoformat()],
    ]
    if output_format == OutputFormat.table.name:
        print_output([], ["Experiment Details"], output_format)
        print_output(table, [], output_format, table_format="plain")
    elif output_format == OutputFormat.json.name:
        print_output(dict(table), [], output_format)


@experiment.command()
@click.option(
    "--experiment-id",
    default=None,
    help="The ID of the experiment to archive, can only supply either an experiment ID or name."
)
@click.option(
    "--experiment-name",
    default=None,
    help="The name of the experiment to archive, can only supply either an experiment ID or name."
)
@click.pass_context
def archive(ctx: click.Context, experiment_id: str, experiment_name: str):
    """Archive an experiment."""
    client = ctx.obj["client"]

    if (experiment_id is None) == (experiment_name is None):
        raise ValueError('Either experiment_id or experiment_name is required')

    if not experiment_id:
        experiment = client.get_experiment(experiment_name=experiment_name)
        experiment_id = experiment.id

    client.archive_experiment(experiment_id=experiment_id)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.pipeline_task."""

import textwrap
import unittest

from absl.testing import parameterized
from kfp import dsl
from kfp.components import pipeline_task
from kfp.components import placeholders
from kfp.components import structures

V2_YAML = textwrap.dedent("""\
components:
  comp-component1:
    executorLabel: exec-component1
    inputDefinitions:
      parameters:
        input1:
          parameterType: STRING
    outputDefinitions:
      artifacts:
        output1:
          artifactType:
            schemaTitle: system.Artifact
            schemaVersion: 0.0.1
deploymentSpec:
  executors:
    exec-component1:
      container:
        args:
        - '{{$.inputs.parameters[''input1'']}}'
        - '{{$.outputs.artifacts[''output1''].path}}'
        command:
        - sh
        - -c
        - echo "$0" >> "$1"
        image: alpine
pipelineInfo:
  name: component1
root:
  dag:
    tasks:
      component1:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-component1
        inputs:
          parameters:
            input1:
              componentInputParameter: input1
        taskInfo:
          name: component1
  inputDefinitions:
    parameters:
      input1:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2
""")


class PipelineTaskTest(parameterized.TestCase):

    def test_create_pipeline_task_valid(self):
        expected_component_spec = structures.ComponentSpec(
            name='component1',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=['sh', '-c', 'echo "$0" >> "$1"'],
                    args=[
                        placeholders.InputValuePlaceholder(
                            input_name='input1')._to_string(),
                        placeholders.OutputPathPlaceholder(
                            output_name='output1')._to_string(),
                    ],
                )),
            inputs={
                'input1': structures.InputSpec(type='String'),
            },
            outputs={
                'output1': structures.OutputSpec(type='system.Artifact@0.0.1'),
            },
        )
        expected_task_spec = structures.TaskSpec(
            name='component1',
            inputs={'input1': 'value'},
            dependent_tasks=[],
            component_ref='component1',
        )
        expected_container_spec = structures.ContainerSpecImplementation(
            image='alpine',
            command=['sh', '-c', 'echo "$0" >> "$1"'],
            args=[
                "{{$.inputs.parameters['input1']}}",
                "{{$.outputs.artifacts['output1'].path}}",
            ],
        )

        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        self.assertEqual(task._task_spec, expected_task_spec)
        self.assertEqual(task.component_spec, expected_component_spec)
        self.assertEqual(task.container_spec, expected_container_spec)

    def test_create_pipeline_task_invalid_wrong_input(self):
        with self.assertRaisesRegex(
                ValueError,
                "Component 'component1' got an unexpected input: 'input0'."):
            task = pipeline_task.PipelineTask(
                component_spec=structures.ComponentSpec.from_yaml_documents(
                    V2_YAML),
                args={
                    'input1': 'value',
                    'input0': 'abc',
                },
            )

    def test_set_caching_options(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_caching_options(False)
        self.assertEqual(False, task._task_spec.enable_caching)

    @parameterized.parameters(
        {
            'cpu': '123',
            'expected_cpu_number': 123,
        },
        {
            'cpu': '123m',
            'expected_cpu_number': 0.123,
        },
        {
            'cpu': '123.0',
            'expected_cpu_number': 123,
        },
        {
            'cpu': '123.0m',
            'expected_cpu_number': 0.123,
        },
    )
    def test_set_valid_cpu_request_limit(self, cpu: str,
                                         expected_cpu_number: float):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_cpu_request(cpu)
        self.assertEqual(expected_cpu_number,
                         task.container_spec.resources.cpu_request)
        task.set_cpu_limit(cpu)
        self.assertEqual(expected_cpu_number,
                         task.container_spec.resources.cpu_limit)

    @parameterized.parameters(
        {
            'gpu_limit': '123',
            'expected_gpu_number': 123,
        },)
    def test_set_valid_gpu_limit(self, gpu_limit: str,
                                 expected_gpu_number: int):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        with self.assertWarnsRegex(
                DeprecationWarning,
                "'set_gpu_limit' is deprecated. Please use 'set_accelerator_limit' instead."
        ):
            task.set_gpu_limit(gpu_limit)
        self.assertEqual(expected_gpu_number,
                         task.container_spec.resources.accelerator_count)

    def test_add_valid_node_selector_constraint(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        with self.assertWarnsRegex(
                DeprecationWarning,
                "'add_node_selector_constraint' is deprecated. Please use 'set_accelerator_type' instead."
        ):
            task.add_node_selector_constraint('TPU_V3')
        self.assertEqual(task.container_spec.resources.accelerator_type,
                         'TPU_V3')

    @parameterized.parameters(
        {
            'limit': '123',
            'expected': 123,
        },
        {
            'limit': 123,
            'expected': 123,
        },
    )
    def test_set_accelerator_limit(self, limit, expected):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )

        task.set_accelerator_limit(limit)
        self.assertEqual(expected,
                         task.container_spec.resources.accelerator_count)

    @parameterized.parameters(
        {
            'memory': '1E',
            'expected_memory_number': 1000000000,
        },
        {
            'memory': '15Ei',
            'expected_memory_number': 17293822569.102703,
        },
        {
            'memory': '2P',
            'expected_memory_number': 2000000,
        },
        {
            'memory': '25Pi',
            'expected_memory_number': 28147497.6710656,
        },
        {
            'memory': '3T',
            'expected_memory_number': 3000,
        },
        {
            'memory': '35Ti',
            'expected_memory_number': 38482.90697216,
        },
        {
            'memory': '4G',
            'expected_memory_number': 4,
        },
        {
            'memory': '45Gi',
            'expected_memory_number': 48.31838208,
        },
        {
            'memory': '5M',
            'expected_memory_number': 0.005,
        },
        {
            'memory': '55Mi',
            'expected_memory_number': 0.05767168,
        },
        {
            'memory': '6K',
            'expected_memory_number': 0.000006,
        },
        {
            'memory': '65Ki',
            'expected_memory_number': 0.00006656,
        },
        {
            'memory': '7000',
            'expected_memory_number': 0.000007,
        },
    )
    def test_set_memory_limit(self, memory: str, expected_memory_number: int):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_memory_request(memory)
        self.assertEqual(expected_memory_number,
                         task.container_spec.resources.memory_request)
        task.set_memory_limit(memory)
        self.assertEqual(expected_memory_number,
                         task.container_spec.resources.memory_limit)

    def test_set_accelerator_type_with_type_only(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_accelerator_type('NVIDIA_TESLA_K80')
        self.assertEqual(
            structures.ResourceSpec(
                accelerator_type='NVIDIA_TESLA_K80', accelerator_count=1),
            task.container_spec.resources)

    def test_set_accelerator_type_with_accelerator_count(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_accelerator_limit('5').set_accelerator_type('TPU_V3')
        self.assertEqual(
            structures.ResourceSpec(
                accelerator_type='TPU_V3', accelerator_count=5),
            task.container_spec.resources)

    def test_set_env_variable(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_env_variable('env_name', 'env_value')
        self.assertEqual({'env_name': 'env_value'}, task.container_spec.env)

    def test_set_display_name(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_display_name('test_name')
        self.assertEqual('test_name', task._task_spec.display_name)

    def test_set_cpu_limit_on_pipeline_should_raise(self):

        @dsl.component
        def comp():
            print('hello')

        @dsl.pipeline
        def graph():
            comp()
            comp()

        with self.assertRaisesRegex(
                ValueError,
                r'set_cpu_limit can only be used on single-step components'):

            @dsl.pipeline
            def my_pipeline():
                graph().set_cpu_limit('1')


class TestPlatformSpecificFunctionality(unittest.TestCase):

    def test_platform_config_to_platform_spec(self):

        @dsl.component
        def comp():
            pass

        @dsl.pipeline
        def my_pipeline():
            t = comp()
            t.platform_config = {'platform1': {'feature': [1, 2, 3]}}
            with self.assertRaisesRegex(
                    ValueError,
                    r"Can only access '\.platform_spec' property on a tasks created from pipelines\. Use '\.platform_config' for tasks created from primitive components\."
            ):
                t.platform_spec


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json
from enum import Enum, unique
from typing import Union

from tabulate import tabulate


@unique
class OutputFormat(Enum):
    """Enumerated class with the allowed output format constants."""
    table = "table"
    json = "json"


def print_output(data: Union[list, dict],
                 headers: list,
                 output_format: str,
                 table_format: str = "simple"):
    """Prints the output from the cli command execution based on the specified
    format.

    Args:
        data (Union[list, dict]): Nested list of values representing the rows to be printed.
        headers (list): List of values representing the column names to be printed
            for the ``data``.
        output_format (str): The desired formatting of the text from the command output.
        table_format (str): The formatting for the table ``output_format``.
            Default value set to ``simple``.

    Returns:
        None: Prints the output.

    Raises:
        NotImplementedError: If the ``output_format`` is unknown.
    """
    if output_format == OutputFormat.table.name:
        click.echo(tabulate(data, headers=headers, tablefmt=table_format))
    elif output_format == OutputFormat.json.name:
        if not headers:
            output = data
        else:
            output = []
            for row in data:
                output.append(dict(zip(headers, row)))
        click.echo(json.dumps(output, indent=4))
    else:
        raise NotImplementedError(
            "Unknown Output Format: {}".format(output_format))



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities functions KFP DSL."""

import re
from typing import Callable, List, Optional, Union

from kfp.deprecated.components import _structures
from kfp.pipeline_spec import pipeline_spec_pb2

_COMPONENT_NAME_PREFIX = 'comp-'
_EXECUTOR_LABEL_PREFIX = 'exec-'

# TODO: Support all declared types in
# components._structures.CommandlineArgumenType
_CommandlineArgumentType = Union[str, int, float,
                                 _structures.InputValuePlaceholder,
                                 _structures.InputPathPlaceholder,
                                 _structures.OutputPathPlaceholder,
                                 _structures.InputUriPlaceholder,
                                 _structures.OutputUriPlaceholder,
                                 _structures.ExecutorInputPlaceholder]


def sanitize_component_name(name: str) -> str:
    """Sanitizes component name."""
    return _COMPONENT_NAME_PREFIX + _sanitize_name(name)


def sanitize_task_name(name: str) -> str:
    """Sanitizes task name."""
    return _sanitize_name(name)


def sanitize_executor_label(label: str) -> str:
    """Sanitizes executor label."""
    return _EXECUTOR_LABEL_PREFIX + _sanitize_name(label)


def _sanitize_name(name: str) -> str:
    """Sanitizes name to comply with IR naming convention.

    The sanitized name contains only lower case alphanumeric characters
    and dashes.
    """
    return re.sub('-+', '-', re.sub('[^-0-9a-z]+', '-',
                                    name.lower())).lstrip('-').rstrip('-')


def get_value(value: Union[str, int, float]) -> pipeline_spec_pb2.Value:
    """Gets pipeline value proto from Python value."""
    result = pipeline_spec_pb2.Value()
    if isinstance(value, str):
        result.string_value = value
    elif isinstance(value, int):
        result.int_value = value
    elif isinstance(value, float):
        result.double_value = value
    else:
        raise TypeError(
            'Got unexpected type %s for value %s. Currently only support str, int '
            'and float.' % (type(value), value))
    return result


# TODO: share with dsl._component_bridge
def _input_artifact_uri_placeholder(input_key: str) -> str:
    return "{{{{$.inputs.artifacts['{}'].uri}}}}".format(input_key)


def _input_artifact_path_placeholder(input_key: str) -> str:
    return "{{{{$.inputs.artifacts['{}'].path}}}}".format(input_key)


def _input_parameter_placeholder(input_key: str) -> str:
    return "{{{{$.inputs.parameters['{}']}}}}".format(input_key)


def _output_artifact_uri_placeholder(output_key: str) -> str:
    return "{{{{$.outputs.artifacts['{}'].uri}}}}".format(output_key)


def _output_artifact_path_placeholder(output_key: str) -> str:
    return "{{{{$.outputs.artifacts['{}'].path}}}}".format(output_key)


def _output_parameter_path_placeholder(output_key: str) -> str:
    return "{{{{$.outputs.parameters['{}'].output_file}}}}".format(output_key)


def _executor_input_placeholder() -> str:
    return "{{{{$}}}}"


def resolve_cmd_lines(cmds: Optional[List[_CommandlineArgumentType]],
                      is_output_parameter: Callable[[str], bool]) -> None:
    """Resolves a list of commands/args."""

    def _resolve_cmd(cmd: Optional[_CommandlineArgumentType]) -> Optional[str]:
        """Resolves a single command line cmd/arg."""
        if cmd is None:
            return None
        elif isinstance(cmd, (str, float, int)):
            return str(cmd)
        elif isinstance(cmd, _structures.InputValuePlaceholder):
            return _input_parameter_placeholder(cmd.input_name)
        elif isinstance(cmd, _structures.InputPathPlaceholder):
            return _input_artifact_path_placeholder(cmd.input_name)
        elif isinstance(cmd, _structures.InputUriPlaceholder):
            return _input_artifact_uri_placeholder(cmd.input_name)
        elif isinstance(cmd, _structures.OutputPathPlaceholder):
            if is_output_parameter(cmd.output_name):
                return _output_parameter_path_placeholder(cmd.output_name)
            else:
                return _output_artifact_path_placeholder(cmd.output_name)
        elif isinstance(cmd, _structures.OutputUriPlaceholder):
            return _output_artifact_uri_placeholder(cmd.output_name)
        elif isinstance(cmd, _structures.ExecutorInputPlaceholder):
            return _executor_input_placeholder()
        else:
            raise TypeError('Got unexpected placeholder type for %s' % cmd)

    if not cmds:
        return
    for idx, cmd in enumerate(cmds):
        cmds[idx] = _resolve_cmd(cmd)



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from typing import Any, Dict, List, Mapping, Optional, Union

import yaml

from .v1_modelbase import ModelBase

PrimitiveTypes = Union[str, int, float, bool]
PrimitiveTypesIncludingNone = Optional[PrimitiveTypes]

TypeSpecType = Union[str, Dict, List]


class InputSpec(ModelBase):
    """Describes the component input specification."""

    def __init__(
        self,
        name: str,
        type: Optional[TypeSpecType] = None,
        description: Optional[str] = None,
        default: Optional[PrimitiveTypes] = None,
        optional: Optional[bool] = False,
        annotations: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(locals())


class OutputSpec(ModelBase):
    """Describes the component output specification."""

    def __init__(
        self,
        name: str,
        type: Optional[TypeSpecType] = None,
        description: Optional[str] = None,
        annotations: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(locals())


class InputValuePlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by the input argument value."""
    _serialized_names = {
        'input_name': 'inputValue',
    }

    def __init__(
        self,
        input_name: str,
    ):
        super().__init__(locals())


class InputPathPlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by a local file path pointing to a file containing the input
    argument value."""
    _serialized_names = {
        'input_name': 'inputPath',
    }

    def __init__(
        self,
        input_name: str,
    ):
        super().__init__(locals())


class OutputPathPlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by a local file path pointing to a file where the program
    should write its output data."""
    _serialized_names = {
        'output_name': 'outputPath',
    }

    def __init__(
        self,
        output_name: str,
    ):
        super().__init__(locals())


class InputUriPlaceholder(ModelBase):  # Non-standard attr names
    """Represents a placeholder for the URI of an input artifact.

    Represents the command-line argument placeholder that will be
    replaced at run-time by the URI of the input artifact argument.
    """
    _serialized_names = {
        'input_name': 'inputUri',
    }

    def __init__(
        self,
        input_name: str,
    ):
        super().__init__(locals())


class OutputUriPlaceholder(ModelBase):  # Non-standard attr names
    """Represents a placeholder for the URI of an output artifact.

    Represents the command-line argument placeholder that will be
    replaced at run-time by a URI of the output artifac where the
    program should write its output data.
    """
    _serialized_names = {
        'output_name': 'outputUri',
    }

    def __init__(
        self,
        output_name: str,
    ):
        super().__init__(locals())


class InputMetadataPlaceholder(ModelBase):  # Non-standard attr names
    """Represents the file path to an input artifact metadata.

    During runtime, this command-line argument placeholder will be
    replaced by the path where the metadata file associated with this
    artifact has been written to. Currently only supported in v2
    components.
    """
    _serialized_names = {
        'input_name': 'inputMetadata',
    }

    def __init__(self, input_name: str):
        super().__init__(locals())


class InputOutputPortNamePlaceholder(ModelBase):  # Non-standard attr names
    """Represents the output port name of an input artifact.

    During compile time, this command-line argument placeholder will be
    replaced by the actual output port name used by the producer task.
    Currently only supported in v2 components.
    """
    _serialized_names = {
        'input_name': 'inputOutputPortName',
    }

    def __init__(self, input_name: str):
        super().__init__(locals())


class OutputMetadataPlaceholder(ModelBase):  # Non-standard attr names
    """Represents the output metadata JSON file location of this task.

    This file will encode the metadata information produced by this task:
    - Artifacts metadata, but not the content of the artifact, and
    - output parameters.

    Only supported in v2 components.
    """
    _serialized_names = {
        'output_metadata': 'outputMetadata',
    }

    def __init__(self, output_metadata: type(None) = None):
        if output_metadata:
            raise RuntimeError(
                'Output metadata placeholder cannot be associated with key')
        super().__init__(locals())

    def to_dict(self) -> Mapping[str, Any]:
        # Override parent implementation. Otherwise it always returns {}.
        return {'outputMetadata': None}


class ExecutorInputPlaceholder(ModelBase):  # Non-standard attr names
    """Represents the serialized ExecutorInput message at runtime.

    This placeholder will be replaced by a serialized
    [ExecutorInput](https://github.com/kubeflow/pipelines/blob/61f9c2c328d245d89c9d9b8c923f24dbbd08cdc9/api/v2alpha1/pipeline_spec.proto#L730)
    proto message at runtime, which includes parameters of the task, artifact
    URIs and metadata.
    """
    _serialized_names = {
        'executor_input': 'executorInput',
    }

    def __init__(self, executor_input: type(None) = None):
        if executor_input:
            raise RuntimeError(
                f'Executor input placeholder cannot be associated with input key. Got {executor_input}'
            )
        super().__init__(locals())

    def to_dict(self) -> Mapping[str, Any]:
        # Override parent implementation. Otherwise it always returns {}.
        return {'executorInput': None}


CommandlineArgumentType = Union[str, InputValuePlaceholder,
                                InputPathPlaceholder, OutputPathPlaceholder,
                                InputUriPlaceholder, OutputUriPlaceholder,
                                InputMetadataPlaceholder,
                                InputOutputPortNamePlaceholder,
                                OutputMetadataPlaceholder,
                                ExecutorInputPlaceholder, 'ConcatPlaceholder',
                                'IfPlaceholder',]


class ConcatPlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by the concatenated values of its items."""
    _serialized_names = {
        'items': 'concat',
    }

    def __init__(
        self,
        items: List[CommandlineArgumentType],
    ):
        super().__init__(locals())


class IsPresentPlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by a boolean value specifying whether the caller has passed an
    argument for the specified optional input."""
    _serialized_names = {
        'input_name': 'isPresent',
    }

    def __init__(
        self,
        input_name: str,
    ):
        super().__init__(locals())


IfConditionArgumentType = Union[bool, str, IsPresentPlaceholder,
                                InputValuePlaceholder]


class IfPlaceholderStructure(ModelBase):  #Non-standard attr names
    '''Used in by the IfPlaceholder - the command-line argument placeholder that will be replaced at run-time by the expanded value of either "then_value" or "else_value" depending on the submissio-time resolved value of the "cond" predicate.'''
    _serialized_names = {
        'condition': 'cond',
        'then_value': 'then',
        'else_value': 'else',
    }

    def __init__(
        self,
        condition: IfConditionArgumentType,
        then_value: Union[CommandlineArgumentType,
                          List[CommandlineArgumentType]],
        else_value: Optional[Union[CommandlineArgumentType,
                                   List[CommandlineArgumentType]]] = None,
    ):
        super().__init__(locals())


class IfPlaceholder(ModelBase):  #Non-standard attr names
    """Represents the command-line argument placeholder that will be replaced
    at run-time by the expanded value of either "then_value" or "else_value"
    depending on the submissio-time resolved value of the "cond" predicate."""
    _serialized_names = {
        'if_structure': 'if',
    }

    def __init__(
        self,
        if_structure: IfPlaceholderStructure,
    ):
        super().__init__(locals())


class ContainerSpec(ModelBase):
    """Describes the container component implementation."""
    _serialized_names = {
        'file_outputs':
            'fileOutputs',  #TODO: rename to something like legacy_unconfigurable_output_paths
    }

    def __init__(
            self,
            image: str,
            command: Optional[List[CommandlineArgumentType]] = None,
            args: Optional[List[CommandlineArgumentType]] = None,
            env: Optional[Mapping[str, str]] = None,
            file_outputs:
        Optional[Mapping[
            str,
            str]] = None,  #TODO: rename to something like legacy_unconfigurable_output_paths
    ):
        super().__init__(locals())


class ContainerImplementation(ModelBase):
    """Represents the container component implementation."""

    def __init__(
        self,
        container: ContainerSpec,
    ):
        super().__init__(locals())


ImplementationType = Union[ContainerImplementation, 'GraphImplementation']


class MetadataSpec(ModelBase):

    def __init__(
        self,
        annotations: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
    ):
        super().__init__(locals())


class ComponentSpec(ModelBase):
    """Component specification.

    Describes the metadata (name, description, annotations and labels),
    the interface (inputs and outputs) and the implementation of the
    component.
    """

    def __init__(
        self,
        name: Optional[str] = None,  #? Move to metadata?
        description: Optional[str] = None,  #? Move to metadata?
        metadata: Optional[MetadataSpec] = None,
        inputs: Optional[List[InputSpec]] = None,
        outputs: Optional[List[OutputSpec]] = None,
        implementation: Optional[ImplementationType] = None,
        version: Optional[str] = 'google.com/cloud/pipelines/component/v1',
        #tags: Optional[Set[str]] = None,
    ):
        super().__init__(locals())
        self._post_init()

    def _post_init(self):
        #Checking input names for uniqueness
        self._inputs_dict = {}
        if self.inputs:
            for input in self.inputs:
                if input.name in self._inputs_dict:
                    raise ValueError(f'Non-unique input name "{input.name}"')
                self._inputs_dict[input.name] = input

        #Checking output names for uniqueness
        self._outputs_dict = {}
        if self.outputs:
            for output in self.outputs:
                if output.name in self._outputs_dict:
                    raise ValueError(f'Non-unique output name "{output.name}"')
                self._outputs_dict[output.name] = output

        if isinstance(self.implementation, ContainerImplementation):
            container = self.implementation.container

            if container.file_outputs:
                for output_name, path in container.file_outputs.items():
                    if output_name not in self._outputs_dict:
                        raise TypeError(
                            'Unconfigurable output entry "{}" references non-existing output.'
                            .format({output_name: path}))

            def verify_arg(arg):
                if arg is None:
                    pass
                elif isinstance(
                        arg, (str, int, float, bool, OutputMetadataPlaceholder,
                              ExecutorInputPlaceholder)):
                    pass
                elif isinstance(arg, list):
                    for arg2 in arg:
                        verify_arg(arg2)
                elif isinstance(
                        arg,
                    (InputUriPlaceholder, InputValuePlaceholder,
                     InputPathPlaceholder, IsPresentPlaceholder,
                     InputMetadataPlaceholder, InputOutputPortNamePlaceholder)):
                    if arg.input_name not in self._inputs_dict:
                        raise TypeError(
                            f'Argument "{arg}" references non-existing input.')
                elif isinstance(arg,
                                (OutputUriPlaceholder, OutputPathPlaceholder)):
                    if arg.output_name not in self._outputs_dict:
                        raise TypeError(
                            f'Argument "{arg}" references non-existing output.')
                elif isinstance(arg, ConcatPlaceholder):
                    for arg2 in arg.items:
                        verify_arg(arg2)
                elif isinstance(arg, IfPlaceholder):
                    verify_arg(arg.if_structure.condition)
                    verify_arg(arg.if_structure.then_value)
                    verify_arg(arg.if_structure.else_value)
                else:
                    raise TypeError(f'Unexpected argument "{arg}"')

            verify_arg(container.command)
            verify_arg(container.args)

        if isinstance(self.implementation, GraphImplementation):
            graph = self.implementation.graph

            if graph.output_values is not None:
                for output_name, argument in graph.output_values.items():
                    if output_name not in self._outputs_dict:
                        raise TypeError(
                            'Graph output argument entry "{}" references non-existing output.'
                            .format({output_name: argument}))

            if graph.tasks is not None:
                for task in graph.tasks.values():
                    if task.arguments is not None:
                        for argument in task.arguments.values():
                            if isinstance(
                                    argument, GraphInputArgument
                            ) and argument.graph_input.input_name not in self._inputs_dict:
                                raise TypeError(
                                    f'Argument "{argument}" references non-existing input.'
                                )

    def save(self, file_path: str):
        """Saves the component definition to file.

        It can be shared online and later loaded using the
        load_component function.
        """

        component_yaml = yaml.dump(self.to_dict(), sort_keys=True)
        with open(file_path, 'w') as f:
            f.write(component_yaml)


class ComponentReference(ModelBase):
    """Component reference.

    Contains information that can be used to locate and load a component
    by name, digest or URL
    """

    def __init__(
        self,
        name: Optional[str] = None,
        digest: Optional[str] = None,
        tag: Optional[str] = None,
        url: Optional[str] = None,
        spec: Optional[ComponentSpec] = None,
    ):
        super().__init__(locals())
        self._post_init()

    def _post_init(self) -> None:
        if not any([self.name, self.digest, self.tag, self.url, self.spec]):
            raise TypeError('Need at least one argument.')


class GraphInputReference(ModelBase):
    """References the input of the graph (the scope is a single graph)."""
    _serialized_names = {
        'input_name': 'inputName',
    }

    def __init__(
            self,
            input_name: str,
            type:
        Optional[
            TypeSpecType] = None,  # Can be used to override the reference data type
    ):
        super().__init__(locals())

    def as_argument(self) -> 'GraphInputArgument':
        return GraphInputArgument(graph_input=self)

    def with_type(self, type_spec: TypeSpecType) -> 'GraphInputReference':
        return GraphInputReference(
            input_name=self.input_name,
            type=type_spec,
        )

    def without_type(self) -> 'GraphInputReference':
        return self.with_type(None)


class GraphInputArgument(ModelBase):
    """Represents the component argument value that comes from the graph
    component input."""
    _serialized_names = {
        'graph_input': 'graphInput',
    }

    def __init__(
        self,
        graph_input: GraphInputReference,
    ):
        super().__init__(locals())


class TaskOutputReference(ModelBase):
    """References the output of some task (the scope is a single graph)."""
    _serialized_names = {
        'task_id': 'taskId',
        'output_name': 'outputName',
    }

    def __init__(
            self,
            output_name: str,
            task_id:
        Optional[
            str] = None,  # Used for linking to the upstream task in serialized component file.
            task:
        Optional[
            'TaskSpec'] = None,  # Used for linking to the upstream task in runtime since Task does not have an ID until inserted into a graph.
            type:
        Optional[
            TypeSpecType] = None,  # Can be used to override the reference data type
    ):
        super().__init__(locals())
        if self.task_id is None and self.task is None:
            raise TypeError('task_id and task cannot be None at the same time.')

    def with_type(self, type_spec: TypeSpecType) -> 'TaskOutputReference':
        return TaskOutputReference(
            output_name=self.output_name,
            task_id=self.task_id,
            task=self.task,
            type=type_spec,
        )

    def without_type(self) -> 'TaskOutputReference':
        return self.with_type(None)


class TaskOutputArgument(ModelBase
                        ):  #Has additional constructor for convenience
    """Represents the component argument value that comes from the output of
    another task."""
    _serialized_names = {
        'task_output': 'taskOutput',
    }

    def __init__(
        self,
        task_output: TaskOutputReference,
    ):
        super().__init__(locals())

    @staticmethod
    def construct(
        task_id: str,
        output_name: str,
    ) -> 'TaskOutputArgument':
        return TaskOutputArgument(
            TaskOutputReference(
                task_id=task_id,
                output_name=output_name,
            ))

    def with_type(self, type_spec: TypeSpecType) -> 'TaskOutputArgument':
        return TaskOutputArgument(
            task_output=self.task_output.with_type(type_spec),)

    def without_type(self) -> 'TaskOutputArgument':
        return self.with_type(None)


ArgumentType = Union[PrimitiveTypes, GraphInputArgument, TaskOutputArgument]


class TwoOperands(ModelBase):

    def __init__(
        self,
        op1: ArgumentType,
        op2: ArgumentType,
    ):
        super().__init__(locals())


class BinaryPredicate(ModelBase):  #abstract base type

    def __init__(self, operands: TwoOperands):
        super().__init__(locals())


class EqualsPredicate(BinaryPredicate):
    """Represents the "equals" comparison predicate."""
    _serialized_names = {'operands': '=='}


class NotEqualsPredicate(BinaryPredicate):
    """Represents the "not equals" comparison predicate."""
    _serialized_names = {'operands': '!='}


class GreaterThanPredicate(BinaryPredicate):
    """Represents the "greater than" comparison predicate."""
    _serialized_names = {'operands': '>'}


class GreaterThanOrEqualPredicate(BinaryPredicate):
    """Represents the "greater than or equal" comparison predicate."""
    _serialized_names = {'operands': '>='}


class LessThenPredicate(BinaryPredicate):
    """Represents the "less than" comparison predicate."""
    _serialized_names = {'operands': '<'}


class LessThenOrEqualPredicate(BinaryPredicate):
    """Represents the "less than or equal" comparison predicate."""
    _serialized_names = {'operands': '<='}


PredicateType = Union[ArgumentType, EqualsPredicate, NotEqualsPredicate,
                      GreaterThanPredicate, GreaterThanOrEqualPredicate,
                      LessThenPredicate, LessThenOrEqualPredicate,
                      'NotPredicate', 'AndPredicate', 'OrPredicate',]


class TwoBooleanOperands(ModelBase):

    def __init__(
        self,
        op1: PredicateType,
        op2: PredicateType,
    ):
        super().__init__(locals())


class NotPredicate(ModelBase):
    """Represents the "not" logical operation."""
    _serialized_names = {'operand': 'not'}

    def __init__(self, operand: PredicateType):
        super().__init__(locals())


class AndPredicate(ModelBase):
    """Represents the "and" logical operation."""
    _serialized_names = {'operands': 'and'}

    def __init__(self, operands: TwoBooleanOperands):
        super().__init__(locals())


class OrPredicate(ModelBase):
    """Represents the "or" logical operation."""
    _serialized_names = {'operands': 'or'}

    def __init__(self, operands: TwoBooleanOperands):
        super().__init__(locals())


class RetryStrategySpec(ModelBase):
    _serialized_names = {
        'max_retries': 'maxRetries',
    }

    def __init__(
        self,
        max_retries: int,
    ):
        super().__init__(locals())


class CachingStrategySpec(ModelBase):
    _serialized_names = {
        'max_cache_staleness': 'maxCacheStaleness',
    }

    def __init__(
            self,
            max_cache_staleness: Optional[
                str] = None,  # RFC3339 compliant duration: P30DT1H22M3S
    ):
        super().__init__(locals())


class ExecutionOptionsSpec(ModelBase):
    _serialized_names = {
        'retry_strategy': 'retryStrategy',
        'caching_strategy': 'cachingStrategy',
    }

    def __init__(
        self,
        retry_strategy: Optional[RetryStrategySpec] = None,
        caching_strategy: Optional[CachingStrategySpec] = None,
    ):
        super().__init__(locals())


class TaskSpec(ModelBase):
    """Task specification.

    Task is a "configured" component - a component supplied with arguments and other applied configuration changes.
    """
    _serialized_names = {
        'component_ref': 'componentRef',
        'is_enabled': 'isEnabled',
        'execution_options': 'executionOptions'
    }

    def __init__(
        self,
        component_ref: ComponentReference,
        arguments: Optional[Mapping[str, ArgumentType]] = None,
        is_enabled: Optional[PredicateType] = None,
        execution_options: Optional[ExecutionOptionsSpec] = None,
        annotations: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(locals())
        #TODO: If component_ref is resolved to component spec, then check that the arguments correspond to the inputs

    def _init_outputs(self):
        #Adding output references to the task
        if self.component_ref.spec is None:
            return
        task_outputs = OrderedDict()
        for output in self.component_ref.spec.outputs or []:
            task_output_ref = TaskOutputReference(
                output_name=output.name,
                task=self,
                type=output.
                type,  # TODO: Resolve type expressions. E.g. type: {TypeOf: Input 1}
            )
            task_output_arg = TaskOutputArgument(task_output=task_output_ref)
            task_outputs[output.name] = task_output_arg

        self.outputs = task_outputs
        if len(task_outputs) == 1:
            self.output = list(task_outputs.values())[0]


class GraphSpec(ModelBase):
    """Describes the graph component implementation.

    It represents a graph of component tasks connected to the upstream
    sources of data using the argument specifications. It also describes
    the sources of graph output values.
    """
    _serialized_names = {
        'output_values': 'outputValues',
    }

    def __init__(
        self,
        tasks: Mapping[str, TaskSpec],
        output_values: Mapping[str, ArgumentType] = None,
    ):
        super().__init__(locals())
        self._post_init()

    def _post_init(self):
        #Checking task output references and preparing the dependency table
        task_dependencies = {}
        for task_id, task in self.tasks.items():
            dependencies = set()
            task_dependencies[task_id] = dependencies
            if task.arguments is not None:
                for argument in task.arguments.values():
                    if isinstance(argument, TaskOutputArgument):
                        dependencies.add(argument.task_output.task_id)
                        if argument.task_output.task_id not in self.tasks:
                            raise TypeError(
                                f'Argument "{argument}" references non-existing task.'
                            )

        #Topologically sorting tasks to detect cycles
        task_dependents = {k: set() for k in task_dependencies.keys()}
        for task_id, dependencies in task_dependencies.items():
            for dependency in dependencies:
                task_dependents[dependency].add(task_id)
        task_number_of_remaining_dependencies = {
            k: len(v) for k, v in task_dependencies.items()
        }
        sorted_tasks = OrderedDict()

        def process_task(task_id):
            if task_number_of_remaining_dependencies[
                    task_id] == 0 and task_id not in sorted_tasks:
                sorted_tasks[task_id] = self.tasks[task_id]
                for dependent_task in task_dependents[task_id]:
                    task_number_of_remaining_dependencies[
                        dependent_task] = task_number_of_remaining_dependencies[
                            dependent_task] - 1
                    process_task(dependent_task)

        for task_id in task_dependencies.keys():
            process_task(task_id)
        if len(sorted_tasks) != len(task_dependencies):
            tasks_with_unsatisfied_dependencies = {
                k: v
                for k, v in task_number_of_remaining_dependencies.items()
                if v > 0
            }
            task_wth_minimal_number_of_unsatisfied_dependencies = min(
                tasks_with_unsatisfied_dependencies.keys(),
                key=lambda task_id: tasks_with_unsatisfied_dependencies[task_id]
            )
            raise ValueError(
                f'Task "{task_wth_minimal_number_of_unsatisfied_dependencies}" has cyclical dependency.'
            )

        self._toposorted_tasks = sorted_tasks


class GraphImplementation(ModelBase):
    """Represents the graph component implementation."""

    def __init__(
        self,
        graph: GraphSpec,
    ):
        super().__init__(locals())


class PipelineRunSpec(ModelBase):
    """The object that can be sent to the backend to start a new Run."""
    _serialized_names = {
        'root_task': 'rootTask',
        #'on_exit_task': 'onExitTask',
    }

    def __init__(
        self,
        root_task: TaskSpec,
        #on_exit_task: Optional[TaskSpec] = None,
    ):
        super().__init__(locals())



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility methods for compiler implementation that is IR-agnostic."""

import collections
from copy import deepcopy
from typing import DefaultDict, Dict, List, Mapping, Set, Tuple, Union

from kfp.components import for_loop
from kfp.components import pipeline_channel
from kfp.components import pipeline_context
from kfp.components import pipeline_task
from kfp.components import tasks_group

GroupOrTaskType = Union[tasks_group.TasksGroup, pipeline_task.PipelineTask]

ILLEGAL_CROSS_DAG_ERROR_PREFIX = 'Illegal task dependency across DSL context managers.'


def additional_input_name_for_pipeline_channel(
        channel_or_name: Union[pipeline_channel.PipelineChannel, str]) -> str:
    """Gets the name for an additional (compiler-injected) input."""

    # Adding a prefix to avoid (reduce chance of) name collision between the
    # original component inputs and the injected input.
    return 'pipelinechannel--' + (
        channel_or_name.full_name if isinstance(
            channel_or_name, pipeline_channel.PipelineChannel) else
        channel_or_name)


def get_all_groups(
    root_group: tasks_group.TasksGroup,) -> List[tasks_group.TasksGroup]:
    """Gets all groups (not including tasks) in a pipeline.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A list of all groups in topological order (parent first).
    """
    all_groups = []

    def _get_all_groups_helper(
        group: tasks_group.TasksGroup,
        all_groups: List[tasks_group.TasksGroup],
    ):
        all_groups.append(group)
        for group in group.groups:
            _get_all_groups_helper(group, all_groups)

    _get_all_groups_helper(root_group, all_groups)
    return all_groups


def get_parent_groups(
    root_group: tasks_group.TasksGroup,
) -> Tuple[Mapping[str, List[str]], Mapping[str, List[str]]]:
    """Get parent groups that contain the specified tasks.

    Each pipeline has a root group. Each group has a list of tasks (leaf)
    and groups.
    This function traverse the tree and get ancestor groups for all tasks.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A tuple. The first item is a mapping of task names to parent groups,
        and second item is a mapping of group names to parent groups.
        A list of parent groups is a list of ancestor groups including the
        task/group itself. The list is sorted in a way that the farthest
        parent group is the first and task/group itself is the last.
    """

    def _get_parent_groups_helper(
        current_groups: List[tasks_group.TasksGroup],
        tasks_to_groups: Dict[str, List[GroupOrTaskType]],
        groups_to_groups: Dict[str, List[GroupOrTaskType]],
    ) -> None:
        root_group = current_groups[-1]
        for group in root_group.groups:

            groups_to_groups[group.name] = [x.name for x in current_groups
                                           ] + [group.name]
            current_groups.append(group)

            _get_parent_groups_helper(
                current_groups=current_groups,
                tasks_to_groups=tasks_to_groups,
                groups_to_groups=groups_to_groups,
            )
            del current_groups[-1]

        for task in root_group.tasks:
            tasks_to_groups[task.name] = [x.name for x in current_groups
                                         ] + [task.name]

    tasks_to_groups = {}
    groups_to_groups = {}
    current_groups = [root_group]

    _get_parent_groups_helper(
        current_groups=current_groups,
        tasks_to_groups=tasks_to_groups,
        groups_to_groups=groups_to_groups,
    )
    return (tasks_to_groups, groups_to_groups)


# TODO: do we really need this?
def get_condition_channels_for_tasks(
    root_group: tasks_group.TasksGroup,
) -> Mapping[str, Set[pipeline_channel.PipelineChannel]]:
    """Gets channels referenced in conditions of tasks' parents.

    Args:
        root_group: The root group of a pipeline.

    Returns:
        A mapping of task name to a set of pipeline channels appeared in its
        parent dsl.Condition groups.
    """
    conditions = collections.defaultdict(set)

    def _get_condition_channels_for_tasks_helper(
        group,
        current_conditions_channels,
    ):
        new_current_conditions_channels = current_conditions_channels
        if isinstance(group, tasks_group.Condition):
            new_current_conditions_channels = list(current_conditions_channels)
            if isinstance(group.condition.left_operand,
                          pipeline_channel.PipelineChannel):
                new_current_conditions_channels.append(
                    group.condition.left_operand)
            if isinstance(group.condition.right_operand,
                          pipeline_channel.PipelineChannel):
                new_current_conditions_channels.append(
                    group.condition.right_operand)
        for task in group.tasks:
            for channel in new_current_conditions_channels:
                conditions[task.name].add(channel)
        for group in group.groups:
            _get_condition_channels_for_tasks_helper(
                group, new_current_conditions_channels)

    _get_condition_channels_for_tasks_helper(root_group, [])
    return conditions


def get_inputs_for_all_groups(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    condition_channels: Mapping[str,
                                Set[pipeline_channel.PipelineParameterChannel]],
    name_to_for_loop_group: Mapping[str, tasks_group.ParallelFor],
) -> Mapping[str, List[Tuple[pipeline_channel.PipelineChannel, str]]]:
    """Get inputs and outputs of each group and op.

    Args:
        pipeline: The instantiated pipeline object.
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_to_parent_groups: The dict of group name to list of
            parent groups.
        condition_channels: The dict of task name to a set of pipeline
            channels referenced by its parent condition groups.
        name_to_for_loop_group: The dict of for loop group name to loop
            group.

    Returns:
        A mapping  with key being the group/task names and values being list
        of tuples (channel, producing_task_name).
        producing_task_name is the name of the task that produces the
        channel. If the channel is a pipeline argument (no producer task),
        then producing_task_name is None.
    """
    inputs = collections.defaultdict(set)

    for task in pipeline.tasks.values():
        # task's inputs and all channels used in conditions for that task are
        # considered.
        task_condition_inputs = list(condition_channels[task.name])

        for channel in task.channel_inputs + task_condition_inputs:

            # If the value is already provided (immediate value), then no
            # need to expose it as input for its parent groups.
            if getattr(channel, 'value', None):
                continue

            # channels_to_add could be a list of PipelineChannels when loop
            # args are involved. Given a nested loops example as follows:
            #
            #  def my_pipeline(loop_parameter: list):
            #       with dsl.ParallelFor(loop_parameter) as item:
            #           with dsl.ParallelFor(item.p_a) as item_p_a:
            #               print_op(item_p_a.q_a)
            #
            # The print_op takes an input of
            # {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item-subvar-q_a;}}.
            # Given this, we calculate the list of PipelineChannels potentially
            # needed by across DAG levels as follows:
            #
            # [{{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item-subvar-q_a}},
            #  {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item}},
            #  {{channel:task=;name=loop_parameter-loop-item-subvar-p_a}},
            #  {{channel:task=;name=loop_parameter-loop-item}},
            #  {{chaenel:task=;name=loop_parameter}}]
            #
            # For the above example, the first loop needs the input of
            # {{channel:task=;name=loop_parameter}},
            # the second loop needs the input of
            # {{channel:task=;name=loop_parameter-loop-item}}
            # and the print_op needs the input of
            # {{channel:task=;name=loop_parameter-loop-item-subvar-p_a-loop-item}}
            #
            # When we traverse a DAG in a top-down direction, we add channels
            # from the end, and pop it out when it's no longer needed by the
            # sub-DAG.
            # When we traverse a DAG in a bottom-up direction, we add
            # channels from the front, and pop it out when it's no longer
            #  needed by the parent DAG.
            channels_to_add = collections.deque()
            channel_to_add = channel

            while isinstance(channel_to_add, (
                    for_loop.LoopArgument,
                    for_loop.LoopArgumentVariable,
            )):
                channels_to_add.append(channel_to_add)
                if isinstance(channel_to_add, for_loop.LoopArgumentVariable):
                    channel_to_add = channel_to_add.loop_argument
                else:
                    channel_to_add = channel_to_add.items_or_pipeline_channel

            if isinstance(channel_to_add, pipeline_channel.PipelineChannel):
                channels_to_add.append(channel_to_add)

            if channel.task_name:
                # The PipelineChannel is produced by a task.

                upstream_task = pipeline.tasks[channel.task_name]
                upstream_groups, downstream_groups = (
                    _get_uncommon_ancestors(
                        task_name_to_parent_groups=task_name_to_parent_groups,
                        group_name_to_parent_groups=group_name_to_parent_groups,
                        task1=upstream_task,
                        task2=task,
                    ))

                for i, group_name in enumerate(downstream_groups):
                    if i == 0:
                        # If it is the first uncommon downstream group, then
                        # the input comes from the first uncommon upstream
                        # group.
                        producer_task = upstream_groups[0]
                    else:
                        # If not the first downstream group, then the input
                        # is passed down from its ancestor groups so the
                        # upstream group is None.
                        producer_task = None

                    inputs[group_name].add((channels_to_add[-1], producer_task))

                    if group_name in name_to_for_loop_group:
                        loop_group = name_to_for_loop_group[group_name]

                        # Pop out the last elements from channels_to_add if it
                        # is found in the current (loop) DAG. Downstreams
                        # would only need the more specific versions for it.
                        if channels_to_add[
                                -1].full_name in loop_group.loop_argument.full_name:
                            channels_to_add.pop()
                            if not channels_to_add:
                                break

            else:
                # The PipelineChannel is not produced by a task. It's either
                # a top-level pipeline input, or a constant value to loop
                # items.

                # TODO: revisit if this is correct.
                if getattr(task, 'is_exit_handler', False):
                    continue

                # For PipelineChannel as a result of constant value used as
                # loop items, we have to go from bottom-up because the
                # PipelineChannel can be originated from the middle a DAG,
                # which is not needed and visible to its parent DAG.
                if isinstance(
                        channel,
                    (for_loop.LoopArgument, for_loop.LoopArgumentVariable
                    )) and channel.is_with_items_loop_argument:
                    for group_name in task_name_to_parent_groups[
                            task.name][::-1]:

                        inputs[group_name].add((channels_to_add[0], None))
                        if group_name in name_to_for_loop_group:
                            # for example:
                            #   loop_group.loop_argument.name = 'loop-item-param-1'
                            #   channel.name = 'loop-item-param-1-subvar-a'
                            loop_group = name_to_for_loop_group[group_name]

                            if channels_to_add[
                                    0].full_name in loop_group.loop_argument.full_name:
                                channels_to_add.popleft()
                                if not channels_to_add:
                                    break
                else:
                    # For PipelineChannel from pipeline input, go top-down
                    # just like we do for PipelineChannel produced by a task.
                    for group_name in task_name_to_parent_groups[task.name]:

                        inputs[group_name].add((channels_to_add[-1], None))
                        if group_name in name_to_for_loop_group:
                            loop_group = name_to_for_loop_group[group_name]

                            if channels_to_add[
                                    -1].full_name in loop_group.loop_argument.full_name:
                                channels_to_add.pop()
                                if not channels_to_add:
                                    break

    return inputs


class InvalidTopologyException(Exception):
    pass


def validate_parallel_for_fan_in_consumption_legal(
    consumer_task_name: str,
    upstream_groups: List[str],
    group_name_to_group: Dict[str, tasks_group.TasksGroup],
) -> None:
    """Checks that a dsl.Collected object is being used results in an
    unambiguous pipeline topology and is therefore legal.

    Args:
        consumer_task_name: The name of the consumer task.
        upstream_groups: The names of the producer task's upstream groups, ordered from outermost group at beginning to producer task at end. This is produced by produced by _get_uncommon_ancestors.
        group_name_to_group: Map of group name to TasksGroup, for fast lookups.
    """
    # handles cases like this:
    # @dsl.pipeline
    # def my_pipeline():
    #     with dsl.ParallelFor([1, 2, 3]) as x:
    #         t = double(num=x)
    #         x = add(dsl.Collected(t.output))
    #
    # and this:
    # @dsl.pipeline
    # def my_pipeline():
    #     t = double(num=1)
    #     x = add(dsl.Collected(t.output))
    producer_task_idx = -1
    producer_task_name = upstream_groups[producer_task_idx]
    if all(group_name_to_group[group_name].group_type !=
           tasks_group.TasksGroupType.FOR_LOOP
           for group_name in upstream_groups[:producer_task_idx]):
        raise InvalidTopologyException(
            f'dsl.{for_loop.Collected.__name__} can only be used to fan-in outputs produced by a task within a dsl.{tasks_group.ParallelFor.__name__} context to a task outside of the dsl.{tasks_group.ParallelFor.__name__} context. Producer task {producer_task_name} is either not in a dsl.{tasks_group.ParallelFor.__name__} context or is only in a dsl.{tasks_group.ParallelFor.__name__} that also contains consumer task {consumer_task_name}.'
        )

    # illegal if the producer has a parent conditional outside of its outermost for loop, since the for loop may or may not be executed
    # for example, what happens if text == 'b'? the resulting execution behavior is ambiguous.
    #
    # @dsl.pipeline
    # def my_pipeline(text: str = ''):
    #     with dsl.Condition(text == 'a'):
    #         with dsl.ParallelFor([1, 2, 3]) as x:
    #             t = double(num=x)
    #     x = add(nums=dsl.Collected(t.output))
    outermost_uncommon_upstream_group = upstream_groups[0]
    group = group_name_to_group[outermost_uncommon_upstream_group]
    if group.group_type in [
            tasks_group.TasksGroupType.CONDITION,
            tasks_group.TasksGroupType.EXIT_HANDLER,
    ]:
        raise InvalidTopologyException(
            f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} When using dsl.{for_loop.Collected.__name__} to fan-in outputs from a task within a dsl.{tasks_group.ParallelFor.__name__} context, the dsl.{tasks_group.ParallelFor.__name__} context manager cannot be nested within a dsl.{group.__class__.__name__} context manager unless the consumer task is too. Task {consumer_task_name} consumes from {producer_task_name} within a dsl.{group.__class__.__name__} context.'
        )
    elif group.group_type != tasks_group.TasksGroupType.FOR_LOOP:
        raise ValueError(
            f'Got unexpected group type when validating fanning-in outputs from task in dsl.{tasks_group.ParallelFor.__name__}: {group.group_type}'
        )


def make_new_channel_for_collected_outputs(
    channel_name: str,
    starting_channel: pipeline_channel.PipelineChannel,
    task_name: str,
) -> pipeline_channel.PipelineChannel:
    """Creates a new PipelineParameterChannel/PipelineArtifactChannel list for
    a dsl.Collected channel from the original task output."""
    if isinstance(starting_channel, pipeline_channel.PipelineParameterChannel):
        return pipeline_channel.PipelineParameterChannel(
            channel_name, channel_type='LIST', task_name=task_name)
    elif isinstance(starting_channel, pipeline_channel.PipelineArtifactChannel):
        return pipeline_channel.PipelineArtifactChannel(
            channel_name,
            channel_type=starting_channel.channel_type,
            task_name=task_name,
            is_artifact_list=True)
    else:
        ValueError(
            f'Got unknown PipelineChannel: {starting_channel!r}. Expected an instance of {pipeline_channel.PipelineArtifactChannel.__name__!r} or {pipeline_channel.PipelineParameterChannel.__name__!r}.'
        )


def get_outputs_for_all_groups(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    all_groups: List[tasks_group.TasksGroup],
    pipeline_outputs_dict: Dict[str, pipeline_channel.PipelineChannel]
) -> Tuple[DefaultDict[str, Dict[str, pipeline_channel.PipelineChannel]], Dict[
        str, pipeline_channel.PipelineChannel]]:
    """Gets a dictionary of all TasksGroup names to an inner dictionary. The
    inner dictionary is TasksGroup output keys to channels corresponding to
    those keys.

    It constructs this dictionary from both data passing within the pipeline body, as well as the outputs returned from the pipeline (e.g., return dsl.Collected(...)).

    Also returns as the second item of tuple the updated pipeline_outputs_dict. This dict is modified so that the values (PipelineChannel) references the group that surfaces the task output, instead of the original task that produced it.
    """

    # unlike inputs, which will be surfaced as component input parameters,
    # consumers of surfaced outputs need to have a reference to what the parent
    # component calls them when they surface them, which will be different than
    # the producer task name and channel name (the information contained in the
    # pipeline channel)
    # for this reason, we use additional_input_name_for_pipeline_channel here
    # to set the name of the surfaced output once

    group_name_to_group = {group.name: group for group in all_groups}
    group_name_to_children = {
        group.name: [group.name for group in group.groups] +
        [task.name for task in group.tasks] for group in all_groups
    }

    outputs = collections.defaultdict(dict)

    # handle dsl.Collected consumed by tasks
    for task in pipeline.tasks.values():
        for channel in task.channel_inputs:
            if not isinstance(channel, for_loop.Collected):
                continue
            producer_task = pipeline.tasks[channel.task_name]
            consumer_task = task

            upstream_groups, downstream_groups = (
                _get_uncommon_ancestors(
                    task_name_to_parent_groups=task_name_to_parent_groups,
                    group_name_to_parent_groups=group_name_to_parent_groups,
                    task1=producer_task,
                    task2=consumer_task,
                ))
            validate_parallel_for_fan_in_consumption_legal(
                consumer_task_name=consumer_task.name,
                upstream_groups=upstream_groups,
                group_name_to_group=group_name_to_group,
            )

            # producer_task's immediate parent group and the name by which
            # to surface the channel
            surfaced_output_name = additional_input_name_for_pipeline_channel(
                channel)

            # the highest-level task group that "consumes" the
            # collected output
            parent_consumer = downstream_groups[0]
            producer_task_name = upstream_groups.pop()

            # process from the upstream groups from the inside out
            for upstream_name in reversed(upstream_groups):
                outputs[upstream_name][
                    surfaced_output_name] = make_new_channel_for_collected_outputs(
                        channel_name=channel.name,
                        starting_channel=channel.output,
                        task_name=producer_task_name,
                    )

                # on each iteration, mutate the channel being consumed so
                # that it references the last parent group surfacer
                channel.name = surfaced_output_name
                channel.task_name = upstream_name

                # for the next iteration, set the consumer to the current
                # surfacer (parent group)
                producer_task_name = upstream_name

                parent_of_current_surfacer = group_name_to_parent_groups[
                    upstream_name][-2]
                if parent_consumer in group_name_to_children[
                        parent_of_current_surfacer]:
                    break

        # handle dsl.Collected returned from pipeline
        for output_key, channel in pipeline_outputs_dict.items():
            if isinstance(channel, for_loop.Collected):
                surfaced_output_name = additional_input_name_for_pipeline_channel(
                    channel)
                upstream_groups = task_name_to_parent_groups[
                    channel.task_name][1:]
                producer_task_name = upstream_groups.pop()
                # process upstream groups from the inside out, until getting to the pipeline level
                for upstream_name in reversed(upstream_groups):
                    new_channel = make_new_channel_for_collected_outputs(
                        channel_name=channel.name,
                        starting_channel=channel.output,
                        task_name=producer_task_name,
                    )

                    # on each iteration, mutate the channel being consumed so
                    # that it references the last parent group surfacer
                    channel.name = surfaced_output_name
                    channel.task_name = upstream_name

                    # for the next iteration, set the consumer to the current
                    # surfacer (parent group)
                    producer_task_name = upstream_name
                    outputs[upstream_name][surfaced_output_name] = new_channel

                # after surfacing from all inner TasksGroup, change the PipelineChannel output to also return from the correct TasksGroup
                pipeline_outputs_dict[
                    output_key] = make_new_channel_for_collected_outputs(
                        channel_name=surfaced_output_name,
                        starting_channel=channel.output,
                        task_name=upstream_name,
                    )
    return outputs, pipeline_outputs_dict


def _get_uncommon_ancestors(
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    task1: GroupOrTaskType,
    task2: GroupOrTaskType,
) -> Tuple[List[GroupOrTaskType], List[GroupOrTaskType]]:
    """Gets the unique ancestors between two tasks.

    For example, task1's ancestor groups are [root, G1, G2, G3, task1],
    task2's ancestor groups are [root, G1, G4, task2], then it returns a
    tuple ([G2, G3, task1], [G4, task2]).

    Args:
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_tor_parent_groups: The dict of group name to list of
            parent groups.
        task1: One of the two tasks.
        task2: The other task.

    Returns:
        A tuple which are lists of uncommon ancestors for each task.
    """
    if task1.name in task_name_to_parent_groups:
        task1_groups = task_name_to_parent_groups[task1.name]
    elif task1.name in group_name_to_parent_groups:
        task1_groups = group_name_to_parent_groups[task1.name]
    else:
        raise ValueError(task1.name + ' does not exist.')

    if task2.name in task_name_to_parent_groups:
        task2_groups = task_name_to_parent_groups[task2.name]
    elif task2.name in group_name_to_parent_groups:
        task2_groups = group_name_to_parent_groups[task2.name]
    else:
        raise ValueError(task2.name + ' does not exist.')

    both_groups = [task1_groups, task2_groups]
    common_groups_len = sum(
        1 for x in zip(*both_groups) if x == (x[0],) * len(x))
    group1 = task1_groups[common_groups_len:]
    group2 = task2_groups[common_groups_len:]
    return (group1, group2)


def get_dependencies(
    pipeline: pipeline_context.Pipeline,
    task_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_parent_groups: Mapping[str, List[str]],
    group_name_to_group: Mapping[str, tasks_group.TasksGroup],
    condition_channels: Dict[str, pipeline_channel.PipelineChannel],
) -> Mapping[str, List[GroupOrTaskType]]:
    """Gets dependent groups and tasks for all tasks and groups.

    Args:
        pipeline: The instantiated pipeline object.
        task_name_to_parent_groups: The dict of task name to list of parent
            groups.
        group_name_to_parent_groups: The dict of group name to list of
            parent groups.
        group_name_to_group: The dict of group name to group.
        condition_channels: The dict of task name to a set of pipeline
            channels referenced by its parent condition groups.

    Returns:
        A Mapping where key is group/task name, value is a list of dependent
        groups/tasks. The dependencies are calculated in the following way:
        if task2 depends on task1, and their ancestors are
        [root, G1, G2, task1] and [root, G1, G3, G4, task2], then G3 is
        dependent on G2. Basically dependency only exists in the first
        uncommon ancesters in their ancesters chain. Only sibling
        groups/tasks can have dependencies.

    Raises:
        RuntimeError: if a task depends on a task inside a condition or loop
            group.
    """
    dependencies = collections.defaultdict(set)
    for task in pipeline.tasks.values():
        upstream_task_names = set()
        task_condition_inputs = list(condition_channels[task.name])
        for channel in task.channel_inputs + task_condition_inputs:
            if channel.task_name:
                upstream_task_names.add(channel.task_name)
        upstream_task_names |= set(task.dependent_tasks)

        for upstream_task_name in upstream_task_names:
            # the dependent op could be either a BaseOp or an opsgroup
            if upstream_task_name in pipeline.tasks:
                upstream_task = pipeline.tasks[upstream_task_name]
            elif upstream_task_name in group_name_to_group:
                upstream_task = group_name_to_group[upstream_task_name]
            else:
                raise ValueError(
                    f'Compiler cannot find task: {upstream_task_name}.')

            upstream_groups, downstream_groups = _get_uncommon_ancestors(
                task_name_to_parent_groups=task_name_to_parent_groups,
                group_name_to_parent_groups=group_name_to_parent_groups,
                task1=upstream_task,
                task2=task,
            )

            # uncommon upstream ancestor check
            uncommon_upstream_groups = deepcopy(upstream_groups)
            uncommon_upstream_groups.remove(
                upstream_task.name
            )  # because a task's `upstream_groups` contains the task's name

            # TODO: this logic can be simplified: the logic to check for consumption from a task in an upstream
            # Condition and ExitHandler should be the same and can be unified with the ParallelFor checks
            if uncommon_upstream_groups:
                dependent_group = group_name_to_group.get(
                    uncommon_upstream_groups[0], None)

                if isinstance(dependent_group,
                              (tasks_group.Condition, tasks_group.ExitHandler)):
                    raise InvalidTopologyException(
                        f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} A downstream task cannot depend on an upstream task within a dsl.{dependent_group.__class__.__name__} context unless the downstream is within that context too. Found task {task.name} which depends on upstream task {upstream_task.name} within an uncommon dsl.{dependent_group.__class__.__name__} context.'
                    )
                elif isinstance(dependent_group, tasks_group.ParallelFor):
                    raise InvalidTopologyException(
                        f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} A downstream task cannot depend on an upstream task within a dsl.{dependent_group.__class__.__name__} context unless the downstream is within that context too or the outputs are begin fanned-in to a list using dsl.{for_loop.Collected.__name__}. Found task {task.name} which depends on upstream task {upstream_task.name} within an uncommon dsl.{dependent_group.__class__.__name__} context.'
                    )

            # tasks in a deeper part of a nested ParallelFor cannot depend directly on a task in a shallower part of the nested ParallelFor
            # to check for this we need to know that:
            # - the upstream task has at least one ParallelFor parent group
            # - the downstream task has at least one ParallelFor parent group uncommon to the upstream task
            # - the downstream consumes from the upstream or has an explicit dependency (.after) on it (a dependency transitively via the parent task group doesn't satisfy the error condition, since the ParallelFor should be permitted to iterate over the upstreams list output)
            if isinstance(upstream_task, pipeline_task.PipelineTask):
                upstream_parent_tasks = task_name_to_parent_groups[
                    upstream_task.name]

                downstream_parallelfor_parents = [
                    group_name_to_group.get(group, None)
                    for group in downstream_groups
                    if isinstance(
                        group_name_to_group.get(group, None),
                        tasks_group.ParallelFor)
                ]
                downstream_in_parallelfor = bool(downstream_parallelfor_parents)

                upstream_parallelfor_parents = [
                    parent_group for parent_group in upstream_parent_tasks
                    if isinstance(
                        group_name_to_group.get(parent_group, None),
                        tasks_group.ParallelFor)
                ]
                upstream_in_parallelfor = bool(upstream_parallelfor_parents)

                downstream_consumes_from_upstream = bool([
                    inp.name
                    for inp in task._task_spec.inputs.values()
                    if isinstance(inp, pipeline_channel.PipelineChannel) and
                    inp.name == upstream_task.name
                ])

                after_called_on_upstream = upstream_task.name in task._run_after

                if downstream_in_parallelfor and upstream_in_parallelfor and (
                        downstream_consumes_from_upstream or
                        after_called_on_upstream):

                    raise InvalidTopologyException(
                        f'{ILLEGAL_CROSS_DAG_ERROR_PREFIX} Downstream tasks in a nested {tasks_group.ParallelFor.__name__} group cannot depend on an upstream task in a shallower {tasks_group.ParallelFor.__name__} group. Task {task.name} depends on upstream task {upstream_task.name}, while {downstream_parallelfor_parents[0]} is nested in {upstream_parallelfor_parents[0]}.'
                    )

            dependencies[downstream_groups[0]].add(upstream_groups[0])

    return dependencies
